#!/usr/bin/env python
# coding: utf-8

"""
Zde leží BlackBox (tuším, bude jeden)
BlackBox pěčlivě ukladá věškerá data,
věškeré sady vzorků, průběžné odhady a tak.
Nejsem už jistý, zda BlackBox je šťastný nazev, neboť
teďkom je to spíše jen krabička pro krámy
"""


import numpy as np
from scipy import spatial
from . import plot
import pickle
from . import IS_stat
from .candybox import CandyBox
from . import sball # for adaptive censoring
from . import f_models # for adaptive censoring
from scipy import stats # for adaptive censoring
from scipy import optimize # for BlackSimpleX

import inspect # for ._log() function

from .samplebox import SampleBox # for candidates packing
from . import simplex as six
from . import lukiskon as lk

from . import estimation as stm # for KechatoLukiskon
import collections #E for Counter in MinEnergyCensoredSampling
            




# considering to rewrite it with pandas
class GuessBox:
    """
    Je tu vynalezaní kola, ale aspoň tak
    Odhady můžou tvořít strašnej bordel a proto jsou
    ukladany do slovníku.
    Tak ale jej nemůžu furt ukladat,
    proto to robim s určitým krokem
    Na konci je třeba zabalit tuhle krabičku ručně!
    """
    def __init__(gb, filename='', flush=100):
        """
        .counter - kdy bylo poslední ukladaní
        """
        gb.estimations = dict()
        gb.filename = filename
        gb.counter = 0
        gb.flush = flush
        gb.pick()
            
    def __repr__(self):
        return "%s(%r, %s)"%(self.__class__.__name__, self.filename, self.flush)
        
    def __str__(gb):
        return str(gb.estimations)
        
    def __len__(gb):
        return len(gb.estimations)
    
    def __getitem__(gb, slice):
        return gb.estimations[slice]
                
    def guess(gb, index, nsim, estimation):
        if index in gb.estimations:
            gb.estimations[index][0].append(nsim)
            gb.estimations[index][1].append(estimation)
        else:
            gb.estimations[index] = [[nsim], [estimation]]
        
        gb.counter+= 1
        if gb.filename and gb.counter > gb.flush:
            gb.put()
        
    def pick(gb):
        if gb.filename:
            try:
                with open(gb.filename + '.pickle', 'rb') as f:
                    gb.estimations = pickle.load(f)
            except:
                # škoda, no
                print("GuessBox: Já tu vaši odhady %s.pickle nevidím" % gb.filename)
        else:
            print('GuessBox is in air mode')
                
    def put(gb):
        if gb.filename:
            try:
                with open(gb.filename + '.pickle', 'wb') as f:
                    pickle.dump(gb.estimations, f)
                gb.counter = 0
            except:
                # nefrčí...
                print("GuessBox: Can not write %s.pickle" % gb.filename)
        else:
            print('GuessBox is in air mode')


class BlackBox:
    """
    BlackBox pěčlivě ukladá věškerá data,
    věškeré sady vzorků (well, no yet), průběžné odhady a tak.
    Nejsem už jistý, zda BlackBox je šťastný nazev, neboť
    teďkom je to spíše jen krabička pro krámy

    .sampled_plan object
    .Z = g_values
    .failsi
    
    Souřadnice primárně z prostoru modelu, ty co jsme rovnou
    posilali do g_modelu!
    
    """
    def __init__(bx, sample_box):
        bx.sample_box = sample_box
        # not really needed
        bx.f = sample_box.sampled_plan()
        # sample density (just helpful thing)
        bx.h = bx.f
        # don't ask me
        bx.candidates = sample_box.sampled_plan()
        
        # nové uložiště odhadů zadám explicitně, aby se pak
        # odhady v stm kodu přířazovaly zprávné krabice
        bx.estimations = []
        # má bejt GuessBox součástí BlackBoxu?
        try:
            bx.guessbox = GuessBox(sample_box.filename, flush=20)
        except:
            bx.guessbox = GuessBox("", flush=20)
        bx.regen()
        
    def __repr__(bx):
        return "%s(%s)"%('BlackBox', repr(bx.sample_box))
        
    def __str__(bx):
        return str('BlackBox ' + bx.sample_box)
        
    def __len__(bx):
        return bx.sample_box.nsim
        
    def __call__(bx):
        """
        Offer next sample
        """
        # I do not see nothing illegal here
        # LHS_like_correction do right conversion
        return bx.LHS_like_correction(bx.h(1))
    
    def __getitem__(bx, slice):
        # stačí vratit sample_box
        return bx.sample_box[slice]
    
    def __getattr__(bx, attr):
        if attr == 'blackbox':
            return bx

        if attr == 'dicebox':
            return bx
            
        # branime sa rekurzii
        # defend against recursion
        # рекурсилы пезьдэт!
        if attr == 'sample_box':
            raise AttributeError
                
        # По всем вопросам обращайтесь 
        # на нашу горячую линию    
        else:
            return getattr(bx.sample_box, attr)
    
    # just plot, green points, red points...
    plot2D = plot.plot2D
    plot3D = plot.plot3D
    show2D = plot.show2D
    show3D = plot.show3D
        
    # přidávání vzorků musí bejt explicitní!
    def add_sample(bx, input_sample):
        bx._log("we have got new data:", str(input_sample))
        bx.sample_box.add_sample(input_sample)
        # tohle musí převest rozdělení vstupního vzorku na vlastní rozdělení skříňky
        inner_sample = bx.sample_box.new_sample(input_sample)
        bx.increment(inner_sample)
        
    
    def increment(bx, input_sample):
        for i in range(bx.nvar):
            for j in range(len(input_sample)):
                plan_index = np.searchsorted(bx.sorted_plan_U[i], input_sample.U[j,i])
                bx.sorted_plan_U[i] = np.insert(bx.sorted_plan_U[i], plan_index, input_sample.U[j,i])
    
    def regen(bx):
        # шайтан регенираци лэзьиз
        bx._logger(msg='regeneration started')
        # pro LHS_like_correction
        bx.sorted_plan_U = [i for i in range(bx.nvar)] # just create list
        for i in range(bx.nvar):
            bx.sorted_plan_U[i] = np.concatenate(([0], np.sort(bx.sampled_plan.U[:, i]), [1]))
            
    # LHS_style correction  
    def LHS_like_correction(bx, input_sample):
        """
        returns sample object (f_model)
        """
        # what is input?
        # as we need transformation anyway,
        # I'll ask for conversion to f sample
        # Здесь вижу железную конвертацию до f-ка,
        # которая пройдёт по R координатам
        # Kruci drát, tady by se nemohlo nic posrat
        to_sample_node = bx.f.new_sample(input_sample)
        
        LHS_node = np.empty(bx.nvar, dtype=float)
        for i in range(bx.nvar):
            if to_sample_node.U.reshape(-1)[i] <= bx.sorted_plan_U[i][0]:
                LHS_node[i] = (bx.sorted_plan_U[i][0] + bx.sorted_plan_U[i][1]) / 2
            elif to_sample_node.U.reshape(-1)[i] >= bx.sorted_plan_U[i][-1]:
                LHS_node[i] = (bx.sorted_plan_U[i][-2] + bx.sorted_plan_U[i][-1]) / 2
            else:
                plan_index = np.searchsorted(bx.sorted_plan_U[i], to_sample_node.U.reshape(-1)[i])
                # vzdy
                LHS_node[i] = (bx.sorted_plan_U[i][plan_index] + bx.sorted_plan_U[i][plan_index - 1]) / 2
                 
        return bx.f.new_sample(LHS_node, 'U')
        
    def _log(bx, *msg, indent=0):
        print(bx.__class__.__name__ + ":", *msg)
            
    def _logi(bx, *msg, indent=1):
        print("\t"*indent, inspect.currentframe().f_back.f_code.co_name + ":", *msg)
        
        
        # The BlackBox Observer 
    def _logger(self, *args, msg="", indent=0, **kwargs):
        print(self.__class__.__name__ + ":" + msg, *args, kwargs) 
        
        
        # inspired by Qt
    def connect(self, slot): self._logger = slot
    def disconnect(self): del(self._logger)
    
    
    # kdyby něco
    # callback = lambda *_, **__: None
    
    
    
    
class Censoring(BlackBox):
    def __init__(bx, sample_object, tri_space='Rn'):
        bx.tri_space = tri_space
        super().__init__(sample_object)
        
    def __repr__(bx):
        return "%s(%s, %s)"%(bx.__class__.__name__, repr(bx.sample_box), repr(bx.tri_space))
        
    def __str__(bx):
        return str(bx.__class__.__name__)
        
    def increment(bx, input_sample):
        super().increment(input_sample)
            
        if "tri" in dir(bx):
            # tri - Deloneho triangulace
            # sample je jíž převeden na f (v .add_sample()), takže je to bezpěčný
            bx.tri.add_points(getattr(input_sample, bx.tri_space))
            # print('increment se podaril')
            if len(bx.tri.coplanar): # pokud triangulace není v pořadku
                #print('triangulace v pořádku není')
                bx._log('triangulation is coplanar')
        else:
            bx._log('Triangulace (zatím?) neexistuje')
            bx.regen()
    
    
    def regen(bx):
        super().regen()
        # chcu zachytit spadnuti QHull na začatku, kdy ještě není dostatek teček.
        #  Jinak je třeba nechat QHull spadnout
        if bx.nsim > 2*bx.nvar + 3: 
            # tady je to OK
            bx.tri = spatial.Delaunay(getattr(bx.sampled_plan, bx.tri_space), incremental=True)
            if len(bx.tri.coplanar):
                #print('triangulace v pořádku není')
                bx._logger(msg='triangulation is coplanar')
            else:
                #print('triangulace je v pořádku')
                bx._logger(msg='triangulation is OK')
            
        else: # lze přípustit chybu triangulace    
            try:
                bx.tri = spatial.Delaunay(getattr(bx.sampled_plan, bx.tri_space), incremental=True)
            except:
                bx._logger(msg='triangulation failed')
        
        
        # pokud tecek nestaci na vytvareni aspon jedneho simplexu - pokrcim rameny
        try:
            # tady je to OK
            bx.tri = spatial.Delaunay(getattr(bx.sampled_plan, bx.tri_space), incremental=True)
            if len(bx.tri.coplanar): # pokud triangulace je v pořadku
                #print('triangulace v pořádku není')
                bx._log('triangulation is coplanar')
            else:
                #print('triangulace je v pořádku')
                bx._log('triangulation is OK')
        except:
            # kdyby neco - berem kramle
            bx._log('triangulation failed')
            
    
    # tato metoda je vlastně pro MinEnergyCensoredSampling
    # ale zde se taky může hodit
    def get_events(bx, simplices=None):
        """
        Metoda musí simplexům přiřazovat jev 
        0=success, 1=failure, 2=mix
        """
        if simplices is None:
            simplices = bx.tri.simplices
        
        in_failure = np.isin(simplices, bx.failure_points)
        has_failure = in_failure.any(axis=1)
        all_failure = in_failure.all(axis=1)
        return np.int8(np.where(has_failure, np.where(all_failure, 1, 2), 0))
    
            
            
    def __call__(bx):
        # je treba si uvedomit ze pravdepodobnost chytnuti muze byt min jak pf. V soucasne realizaci
        try:
            # očekávám, že projde kandidaty
            # vodkaď jsou - netuším
            to_sample, rate = bx.filter()
            # když nepovedlo s kandidaty, zkusíme sami nagenerovat
            while len(to_sample) == 0: # nekonečno
                # tak, děcka, server má spoustu času a nikam nespejchá
                # ale pokud tohle sežere věškerou paměť (ne když, ale kdy)
                # dostaneš co proto!
                # vomezíme pole na 10e6 (desitky mega teda vodhadově)
                # by bylo možně použit chytrejší vzoreček s logaritmem
                # ale snad tohle postačí
                to_sample, rate = bx.filter(bx.h(int(0.9999*rate + 100)))
            return bx.LHS_like_correction(to_sample)
        
        # chcu zachytit spadnuti QHull na začatku, kdy ještě není
        # dostatek teček. Je-li bx.tri fakticky existuje, tj.
        # triangulace jíž existovala - je třeba nechat QHull spadnout
        except AttributeError:
            # kdyby neco - berem kramle
            print("Triangulation doesn't exist. Censoring failed")
            return bx.LHS_like_correction(bx.h(1)) # it should be OK
           
    def filter(bx, candidates=[]):
        """
        supports both sample and sample.R input
        
        logika metody, nebo, přesněji, implementaci 
        je taková, že bule-li někdo-něco mimo doménu,
        tak funkce je vrátí a zbytek už neřeší
        """
        # co to bylo na vstupu?
        # když nebyl žádný,
        # projdeme vlastními kandidaty
        if len(candidates)==0:
            candidates = bx.candidates
            
        
        # tady byl problém. Funkce byla původně navržena tak, 
        # aby ji nezajimalo co je na vstupu
        # to ale nefunguje
        # další funkce jako výstup očekavají něco s validním R-kem
        candidates_to_sample_node = getattr(bx.f.new_sample(candidates), bx.tri_space)
            
            
        found_simplices = bx.tri.find_simplex(candidates_to_sample_node)
        # ouside of domain - it's easy
        outside = candidates[found_simplices < 0]
        if len(outside) > 0:
            # my hodnotili svých kandidatov?
            if bx.candidates == candidates:
                bx.candidates = outside
            return outside, len(candidates)/len(outside)
            
        # tady já chcu vrátit první vhodný vzorek a tím končít
        for candidate_id in range(len(candidates)):
            # simplex_id >= 0: # inside of domain
            simplex_id = found_simplices[candidate_id]
            simplex = bx.tri.simplices[simplex_id]

            # fp like a failure points. Number of failure points
            fp = len(np.setdiff1d(simplex, bx.failure_points))

            # pokud je simplex není jednobarevny..
            if (fp != 0) and (fp != bx.nvar+1):
                # my hodnotili svých kandidatov?
                if bx.candidates == candidates:
                    bx.candidates = candidates[candidate_id:]
                return candidates[candidate_id], candidate_id
            
        # nepovedlo. nic
        # mě nenapadá žádný lepší způsob vrátit prázdnou matici
        return candidates[0:0], len(candidates)
           
           
           
           
class AdaptiveCensoring(Censoring):
    def __init__(bx, sample_object, tri_space='Rn', pf_lim=(1,0)):
        bx._logger(msg="instance creating")
        bx.sball = sball.Sball(sample_object.nvar)
        bx.pf_lim = pf_lim
        
        bx.base_r = bx.sball.get_r(0.5)
        
        # pro jistotu pridame
        bx.simplex_index = {'failure':[], 'success':[], 'mix':[]}
        
        
        # overall estimations
        #bx.oiss = IS_stat.ISSI(['failure', 'success', 'out', 'mix'])
        # -1 = 'out', 0=success, 1=failure, 2=mix
        bx.oiss = IS_stat.ISSI([-1,0,1,2]) 
        # current estimations
        bx.ciss = IS_stat.ISSI([-1,0,1,2]) 
        
        super().__init__(sample_object, tri_space)
        
        
    def __repr__(bx):
        return "%s(%s, %s, %s)"%(bx.__class__.__name__, repr(bx.sample_box), repr(bx.tri_space), repr(bx.pf_lim))
        
    def __str__(bx):
        return str(bx.__class__.__name__)
        
    def regen(bx):
        bx._logi(inspect.currentframe().f_back.f_code.co_name, "launched regeneration")
            
        super().regen()
        if bx.pf_lim[1] == 0:
            bx.drop_r = float("inf")
        else:
            bx.drop_r = bx.sball.get_r(bx.pf_lim[1])
        
        # dropneme (pro jistotu) odhady
        bx.oiss = bx.ciss
        # -1 = 'out', 0=success, 1=failure, 2=mix
        bx.ciss = IS_stat.ISSI([-1,0,1,2]) 
        
        # drop indexes
        bx.simplex_index = {'failure':[], 'success':[], 'mix':[]}
        
    def increment(bx, input_sample):
        super().increment(input_sample)
        
        # drop indexes
        bx.simplex_index = {'failure':[], 'success':[], 'mix':[]}
        
        # current estimations
        try: # to čo já vidím v kódu - ISSI slovníky se pokažde generujóu znovu, 
            # není nutně je explicitně kopirovať
            bx.guessbox.guess('TRI_overall_estimations', bx.nsim-1, bx.oiss.estimations)
            bx.guessbox.guess('TRI_current_estimations', bx.nsim-1, bx.ciss.estimations)
        except AttributeError:
            bx.guessbox.guess('TRI_upper_pf', bx.nsim-1, 1)
            
        # a znovu začneme počítat
        # -1 = 'out', 0=success, 1=failure, 2=mix
        bx.ciss = IS_stat.ISSI([-1,0,1,2]) 
            
            
    def __call__(bx):
        bx._logger(msg="we were asked for an recommendation")
        # je treba si uvedomit ze pravdepodobnost chytnuti muze byt min jak pf. V soucasne realizaci
        try:
            # očekávám, že projde kandidaty
            # odkaď jsou - netuším
            to_sample, rate, simplex_id = bx.filter()
            # když nepovedlo s kandidaty, zkusíme sami nagenerovat
            while len(to_sample) == 0: # nekonečno
                # pokusme se nastavit rate tak, abychom získali právě jedneho kandidata
                try: # try uvnitř traja
                    p_rate = bx.oiss.estimations[-1] + bx.oiss.estimations[2]
                except:
                    p_rate = bx.pf_lim[0] - bx.pf_lim[1]
                if p_rate < 1e-5:
                    rate = 100000
                else:
                    rate = int(1/p_rate) + 1
                to_sample, rate, simplex_id = bx.filter(bx.get_candidates(rate))
            choose = bx.LHS_like_correction(to_sample)
            bx._log("finally we choose", str(choose), "of", simplex_id, "simplex")
            return choose
        
        # chcu zachytit spadnuti QHull na začatku, kdy ještě není
        # dostatek teček. Je-li bx.tri fakticky existuje, tj.
        # triangulace jíž existovala - je třeba nechat QHull spadnout
        except AttributeError:
            choose = bx.LHS_like_correction(bx.get_candidates(1))
            if bx.nsim < bx.nvar + 1: # je to legální
                bx._log("we have no enough points to build triangulation, so", str(choose), "is our recommendation")
                return choose
                
            elif bx.nsim < 2*bx.nvar + 3: # to je ještě budiž
                bx._log("we have troubles with triangulation, so we offer random sample for now:", str(choose))
                return choose
            else: # no to teda ne!
                raise ValueError("AdaptiveCensoring: s tou triangulací je fakt něco není v pořadku")
           
           
    def filter(bx, candidates=[]):
        """
        za pvré, jako vstup očekávám kandidaty od .get_candidates() funkce,
        zabalené do полукустарного sample_boxu s zadaným .implicit_multiplicator
        (je to drobnost pro přesnějši zpracování sad IS IS_statem).
        
        Metoda musí souřádnicím přiřazovat jev 
        "success", "failure", "mix", "outside"
        
        TATO metoda jakmile narazí na "mix" nebo "outside"
        ukladá zjištěné informace do ISSI a nalezeného kandidata vrací
        """
        # co to bylo na vstupu?
        # když nebyl žádný,
        # projdeme vlastními kandidaty
        if len(candidates)==0:
            candidates = bx.candidates
            
        bx._logi("kandidaty:", candidates)
            
        # je třeba lokálně zachovat implicit_multiplicator
        # jinak se ztrací při slajsingu
        # nechceš přepsat SampleBox, Alexi?   
        try:     
            implicit_multiplicator = candidates.implicit_multiplicator
        except AttributeError: # kandidaty můžou bejt odkudkoliv
            # s nekonečným rozptylem nebudou mít váhu "absenční" jevy
            # moc to odhadům nepomůže, protože je-li kandidaty
            # nemajú .implicit_multiplicator
            # asi nebudou mať ani váhy IS v .g_values
            implicit_multiplicator = float("inf")
            bx._logi("Dobrý den, kandidaty nemajú .implicit_multiplicator")#. S pozdravem, AdaptiveCensoring")
        
        # tady byl problém. Funkce byla původně navržena tak, 
        # aby ji nezajimalo co je na vstupu
        # to ale nefunguje
        # další funkce jako výstup očekavají něco s validním R-kem
        # no tj. já zde provádím posouzení transformací z R-ka vstupních souřadnic
        candidates_to_sample_node = getattr(bx.f.new_sample(candidates), bx.tri_space)
            
            
        current_simplices = bx.tri.find_simplex(candidates_to_sample_node)
        
        
        # tak bacha
        # budeme přepísovat jevy in-place
        found_simplices = np.ma.array(current_simplices.copy()) #.copy()
        # nemaskované - obsahuji číslo simplexu
        # maskované - číslo jevu
        # -1 = 'out', 0=success, 1=failure, 2=mix
        # tj. procházíme simplexy z náhodné sady vzorků,
        # nahrazujeme čislo simplexu odpovidajicím mu jevem
        # a skryváme ho
        # pote ty "skryté", "projduté" vzorky využiváme k žískání odhadů
        
        while len(current_simplices):# > 0:
            bx._logi("current simplices", current_simplices)
            # berem hned prvního kandidata 
            # a posuzujeme co je zač
            simplex_id = current_simplices[0]
            mask = found_simplices==simplex_id
            
            if simplex_id < 0: # -1 means ouside
                # berem kramle
                break
            elif simplex_id in bx.simplex_index['success']:
                found_simplices[mask] = 0
            elif simplex_id in bx.simplex_index['failure']:
                found_simplices[mask] = 1
            elif simplex_id in bx.simplex_index['mix']:
                found_simplices[mask] = 2
                # kramle
                break
            else: # no index information
                # tady já chcu vrátit první vhodný vzorek a tím končít
                # simplex_id >= 0: # inside of domain
                # asi tady získavam množinu s čísly vrcholů
                # kteří zakladají simplex
                simplex = bx.tri.simplices[simplex_id]
                
                # for debug
                bx._logi("провал индексу", simplex_id, indent=2)
    
                # fp like a failure points. Number of failure points
                # setdiff "Return the unique values in ar1 that are not in ar2."
                fp = len(np.setdiff1d(simplex, bx.failure_points))
                
                if fp == bx.nvar+1: # 
                    bx.simplex_index['success'].append(simplex_id)
                    found_simplices[mask] = 0
                elif fp == 0:
                    bx.simplex_index['failure'].append(simplex_id)
                    found_simplices[mask] = 1
                    #print("failure simplex", simplex_id)
                else:
                    bx.simplex_index['mix'].append(simplex_id)
                    found_simplices[mask] = 2
                    bx._logi("mixed simplex", simplex_id)
                    # bacha! kramle
                    break
    
            # pridame do seznamu známého
            found_simplices[mask] = np.ma.masked
            # eště raz
            cmask = current_simplices==simplex_id
            # vyfiltrujeme
            current_simplices = current_simplices[~cmask]
    
    
            
        # zde je třeba перехватить ситуацию, куке одӥг но кандидат ӧвӧл
        # нужно ли? 
#        if len(current_simplices) == 0:
#            # nepovedlo. nic
#            bx.candidates = candidates[0:0]
#            # mě nenapadá žádný lepší způsob vrátit prázdnou matici
#            return candidates[0:0], len(candidates), -2 # simple_id
                
        # nemaskované, včetně současného kandidata (nevím proč) - ke kandidatům
        # землю - крестьянам, фабрики - рабочим
        # předpokladam, že kandidaty jsou se všim všudy
        # s vahami (.g_value) a se svým .implicit_multiplicator'em
        ## zde True hodnoty roušky - to co jíž bylo skryto
        bx.candidates = candidates[~np.ma.getmaskarray(found_simplices)][1:] # toho prvního prečo nechcem
        try: # na zacatku je tam prazdný f_model, kterému atribut pripsat nemůžeme
            bx.candidates.implicit_multiplicator = implicit_multiplicator
        except:
            pass
        
        # vrátíme kandidaty, všechny-ne všechny?
        # малы одӥг гинэ? уг тодӥськы чик...
        selected_candidate = candidates[~np.ma.getmaskarray(found_simplices)][:1] # chcem toho prvního
        
        
        # odešleme ISSI
        try:
            # pridame do seznamu známého
            # rouška musí zůstat z cyklu
            # proboha, Alexi, co to je za roušku, co se tu děje?
            # jakmile v tom hlavním cyklu nalezli jsme mix nebo outside
            # my hned z cyklu vylezli a je neskryli - abychom je vzali jako kandidaty
            # teď je však skryváme s tou "rouškou", co musela být před opuštěním cyklu nastavena
            # tak ISSI bude mít možnost odhadovat i pravděpodobnosti mix a outside
            found_simplices[mask] = np.ma.masked
            # zde získáme True hodnoty roušek
            # ukazatel
            imask = found_simplices.mask
            found_simplices.mask = ~imask # invertujem, dotkne to i samotnou imask
            events = found_simplices.compressed()
            print(candidates)
            print(candidates.g_values)
            print("imask", imask)
            print(candidates.g_values[~imask])
            print(events)
            bx.oiss.add_IS_serie(candidates.g_values[~imask], events, implicit_multiplicator)
            print("global estimation", bx.oiss.estimations)
            bx.ciss.add_IS_serie(candidates.g_values[~imask], events, implicit_multiplicator)
            print("current estimation", bx.ciss.estimations)
        except AttributeError:
            bx._logi("Это вы мне прислали неваженых кандидатов?")
        except UnboundLocalError: # čo tu chybu způsobuje?
            # asi nebyly žádné kandidaty (třeba hned na začátku)
            assert len(candidates)==0 and len(bx.candidates)==0, "AdaptiveCensoring: Что за бурда с этими кандидатама?"
            return candidates, 0, -2
            
        
        # rate = kolík bylo - kolik zůstalo
        return selected_candidate, len(candidates) - len(bx.candidates), simplex_id
            
       
    
    
    def get_candidates(bx, Nsim=int(1e4)):
        # -1 = 'out', 0=success, 1=failure, 2=mix
        
        # не мудрствуя лукаво
        user_pf = np.mean(bx.pf_lim)
        try:
            low_pf = bx.oiss.estimations[1] # failure
            upper_pf = 1 - bx.ciss.estimations[0] # sucess
            self_pf = (low_pf + upper_pf)/2
        except AttributeError:
            self_pf = 0.5
            
            # bereme *mean* od svého a uživatelského odhadu
            # minimum nejede
        sampling_r, __ = bx.sball.get_r_iteration(np.mean((self_pf, user_pf)))
        # asi tam bylo sampling_r/bx.base_r, že?
        # u stats.norm zadáváme směrodatnou odchylku, je to asi správné
        h = f_models.UnCorD([stats.norm(0, sampling_r/bx.base_r) for i in range(bx.nvar)])
        
        # for IS_stats
        svar = (sampling_r/bx.base_r)**2 # svar like sampling_variance
        # něco takovýho bych nahrubo placnul
        implicit_multiplicator = svar**bx.nvar * np.exp(bx.nvar/svar - bx.nvar)
            
        #
        # jdeme na to, koťě!
        #
        
        # zgenerujeme vzorky
        # nic zajimavýho
        h = h(Nsim)
        # dropneme priliš vzdálené kandidaty
        distance_from_zero = np.sum(h.R**2, axis=1)
        mask = distance_from_zero < bx.drop_r
        
        # a teď bacha!
        # tady musíme provést jeden trik
        to_sample = bx.f.new_sample(h.R[mask], 'G') # R-ko smerdžíme ako G-čko
        w = to_sample.pdf_G / h.pdf_R # snad je to správně
        # zabalme do boxu
        candidates = SampleBox(to_sample, w, 'BlackBox internal samples and weights')
        candidates.implicit_multiplicator = implicit_multiplicator
        # vahy máme, zbytek už nejsou naši starosti
        return candidates
       






class MinEnergyCensoredSampling(Censoring):
    # už mě to dědění nebaví
    # без поллитры не разберёшься, что этот слоёный пирог делает
    def __init__(bx, sample_object, tri_space='Rn', tree_space=None, sampling_space=None, kechato_space='U', \
                                        potencial='ksee', p_norm=2, budget=1000, simplex_budget=100):
        
        bx.sball = sball.Sball(sample_object.nvar)
        bx.base_r = bx.sball.get_r(0.5)
        
        if tree_space is None:
            bx.tree_space = tri_space
        else:
            bx.tree_space = tree_space
            
        if sampling_space is None:
            bx.sampling_space = tri_space
        else:
            bx.sampling_space = sampling_space
        
        bx.kechato_space = kechato_space
        # pro simplexy. Chcu ukladat jejich míry
        # viz. .regen()
        #bx.siss = IS_stat.ISSI()
        bx.budget = budget
        bx.simplex_budget = simplex_budget
        bx.p_norm = p_norm
        bx.potencial = potencial
        
        # for current candidates
        # kandidaty musí být 'judged' a 'assessed'
        # viz. regen()
        #bx.candidates_index = dict()
        # krám, přece třidíme odpad!
        bx.former_candidates = []
        bx.unjudged_candidates = []
        # user .candidates should be initializated in the base class
        # but we need to set additional attributes to them
        super().__init__(sample_object, tri_space)
        bx.candidates = CandyBox(bx.f())
        
        
        
        
    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        super().regen()
        # zás mám vyhodit odhady?
        bx.siss = IS_stat.ISSI() 
        bx.candidates_index = dict()
        
        sampled_plan_tree = getattr(bx.sample_box, bx.tree_space)
        bx.tree = spatial.cKDTree(sampled_plan_tree)
        
        if "tri" in dir(bx):
            # tri - Deloneho triangulace
            bx.simplex_events = bx.get_events()
            bx.estimate_outside()
            for simplex_id in range(bx.tri.nsimplex):
                # zde jen počítame
                bx.estimate_simplex(simplex_id)
                
                
        elif bx.nsim > 0: # požaduji, aby nějaké těčíčky byly vždy pritomné
            
            candidates = IS_stat.IS(bx.f, bx.h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
            # nevím co tam bylo za h-ko, ale nechť IM zůstane 1
            implicit_multiplicator = 1
            
            # teoreticky lze
            #bx.siss.add_IS_serie(candidates.w, candidates.simplex, implicit_multiplicator)
        
            bx.assess_candidates(candidates)
            # uložíme
            bx.candidates_index[-1] = candidates
            
            
            
        
    
    
    def export_estimation(bx):
        bx.siss.get_estimations()
        simplices = np.array(tuple(bx.siss.estimations.keys()))
        probabilities = np.array(tuple(bx.siss.estimations.values()))
        
        estimation = dict()
        estimation[-1] = np.sum(probabilities[simplices == -1])
        
        # jevy aj klidně in-place (nerobím kopiju)
        events = simplices[simplices != -1]
        probabilities = probabilities[simplices != -1]
        
        # zhruba - get_events() vrací pole s odpovidajícími čísly jevů pro každý simplex, počineje od nuly
        # tím slajsingem my jakoby vybirame ke každemu nalezenemu simplexovi ten správnej mu odpovídajicí jev
        events = bx.simplex_events[events]
        
        for i in range(3): # kvůli 0,1,2 robiť cyklus?
            estimation[i] = np.sum(probabilities[events == i])
        
        bx.guessbox.guess('TRI_overall_estimations', bx.tri.npoints, estimation)
        
        
    def increment(bx, input_sample):
        # ну нахрен это ваше наследование-расследование
        
        # nechť bude, asi nikomu nevadí
        for i in range(bx.nvar):
            for j in range(len(input_sample)):
                plan_index = np.searchsorted(bx.sorted_plan_U[i], input_sample.U[j,i])
                bx.sorted_plan_U[i] = np.insert(bx.sorted_plan_U[i], plan_index, input_sample.U[j,i])
        
        #č strom posuneme sem    
        sampled_plan_tree = getattr(bx.sample_box, bx.tree_space)
        bx.tree = spatial.cKDTree(sampled_plan_tree)
        
        
        # fór je v tom, že tu triangulaci nemůžeme výtvořit hned na začátku
        if "tri" in dir(bx):
            # tri - Deloneho triangulace
            
            bx.export_estimation()
            
            # чыры-пыры
            # invalidujeme staré vzorky
            if -2 in bx.candidates_index:
                bx.former_candidates.append(bx.candidates_index.pop(-2))
            
            
            former_simplices = bx.tri.simplices
            mixed = bx.is_mixed()
            
            # sample je jíž převeden na f (v .add_sample()), takže je to bezpěčný
            bx.tri.add_points(getattr(input_sample, bx.tri_space))
            bx.simplex_events = bx.get_events()
            # print('increment se podaril')
            if len(bx.tri.coplanar): # pokud triangulace není v pořadku
                #print('triangulace v pořádku není')
                bx._logger(msg='triangulation has coplanar points')
                
            
            
            
            # zkontrolujeme co se změnilo
            # předpokladám, že se počet simplexů přidaním bodů nezměnší 
            equal_mask = former_simplices == bx.tri.simplices[:len(former_simplices)]
            changed_simplices_ids = np.argwhere(~equal_mask.all(axis=1)).reshape(-1)
            
            # invalidirujeme jejich odhady
            for simplex_id in changed_simplices_ids:
                bx.siss.delete_event_data(simplex_id)
                
                # popajem pouze mixy, ty musel jsem spočítat před aktualizací
            for simplex_id in changed_simplices_ids[mixed[changed_simplices_ids]]:
                bx.former_candidates.append(bx.candidates_index.pop(simplex_id))
                
            
            # pokud není splněná podmínka, 
            # tak nemáme jistotu, že se potenciály nezměni
            # ani u kandidatů, které se nacházejí v pojíštěných státem buňkách
            if (bx.tree_space != bx.tri_space) or (bx.p_norm != 2):
                for candidates in bx.candidates_index.values():
                    bx.assess_candidates(candidates)    
                
            # změněné simplexy přepočítáme
            for simplex_id in changed_simplices_ids:
                bx.estimate_simplex(simplex_id)    
                
                
                
            # teď nové simplexy
            # simplexy свежего разлива
            for simplex_id in range(len(former_simplices), bx.tri.nsimplex):
                # zde jen počítame
                bx.estimate_simplex(simplex_id)
            
            
            
            # přepočíst -1 v zavislosti na simplexu vstupního bodu
            try:
                # kontrola korrektní i v případě NaN
                test = input_sample.simplex > -1
                # эскером
                if not test.all():
                    bx.former_candidates.append(bx.candidates_index.pop(-1))
                    bx.siss.delete_event_data(-1)
                    bx.estimate_outside()
            except BaseException as e:
                msg = "input sample didn't provide correct 'simplex' attribute "
                error_msg = bx.__class__.__name__ + ": " + msg + repr(e)
                bx._logger(msg=error_msg)
                bx.former_candidates.append(bx.candidates_index.pop(-1))
                bx.siss.delete_event_data(-1)
                bx.estimate_outside()
                
                
            
            
            # pokud -2 jíž existuje, tak pro dnešek stačí
            if (-2 not in bx.candidates_index) and (len(bx.former_candidates) > 0):
                # nikdo nám neslibil, že u starých kandidatu 
                # třeba se nebude zvýšovat potanciál
                # (je to prostě, opravdu ríct, jednodušší)
                
                # prohrabeme odpad
                candidates = bx.former_candidates.pop()
                for i in range(len(bx.former_candidates)):
                    candidates.add_sample(bx.former_candidates.pop())
                for i in range(len(bx.unjudged_candidates)):
                    candidates.add_sample(bx.unjudged_candidates.pop())
                
                # hodil by se ještě nám?
                bx.judge_candidates(candidates)
                # profiltrujeme
                # -1 = 'out', 0=success, 1=failure, 2=mix
                candidates = candidates[candidates.event != 0] 
                candidates = candidates[candidates.event != 1] 
                # uvalíme pokutu
                bx.assess_candidates(candidates)
                #č někoho z tyto hromady dostaneme 
                for __ in range((bx.nvar+1)*2):
                    if len(candidates) > 0:
                        mc_id = collections.Counter(candidates.simplex).most_common()[0][0]
                        #č uložíme
                        bx.candidates_index[mc_id].add_sample(candidates[candidates.simplex == mc_id])
                        #č a teď bacha - ty co jsem uložil do simplexu nechcu v -2 videt
                        candidates = candidates[candidates.simplex != mc_id]
                    else:
                        break
                if len(candidates) > 0:
                    #č -2 je určen pro zbytky, кылем-мылем
                    bx.candidates_index[-2] = candidates
                
                
        else:
            bx._logger('Triangulace (zatím?) neexistuje')
            bx.regen()
        
        
    def is_mixed(bx, simplices=None):
        """
        Metoda musí simplexům přiřazovat jev 
        0=success, 1=failure, 2=mix
        """
        if simplices is None:
            simplices = bx.tri.simplices
        
        in_failure = np.isin(simplices, bx.failure_points)
        has_failure = in_failure.any(axis=1)
        all_failure = in_failure.all(axis=1)
        return np.logical_xor(has_failure, all_failure)
            
            
    def __call__(bx):
        bx._logger(msg="we were asked for an recommendation")
        
        if bx.nsim < 1: # je to legální
            bx._logger(msg="median first!")
            return bx.LHS_like_correction(bx.h(1))
        else:
            selected = bx.select_candidate()
            # hrubě ošidíme s takovejhlema usilima námi pěčlivě vybraný vzorečiček
            #selected.sampling_plan = bx.LHS_like_correction(selected)
            return selected
           
    
    
    def select_candidate(bx):
    
        # 
        # AUKCE, AUCTION    
        # 
        # for current candidates
        # kandidaty musí být 'judged' a 'assessed'
        highest_bid = 0 
        for candidates in bx.candidates_index.values():
            bids = getattr(candidates, bx.potencial)
            bid = np.max(bids)
            # side effect
            if bid > highest_bid:
                bidder = candidates[np.argmax(bids)]
                highest_bid = bid
        
        
        
        # probably, we shouldn't purge user candidates (if they are)
        # just every time evaluate them
        if len(bx.candidates) > 0:
            bx.judge_candidates(bx.candidates)
            bx.assess_candidates(bx.candidates)
            bids = getattr(bx.candidates, bx.potencial)
            # -1 = 'out', 0=success, 1=failure, 2=mix
            bids *= (bx.candidates.event == -1) + (bx.candidates.event == 2)
            bid = max(bids)
            # side effect
            if bid > highest_bid:
                bidder = bx.candidates[np.argmax(bids)]
                # ... кулэ ӧвӧл
        
        
        # já prečo spoléhám na ksee a psee potenciály
        # v tom, že nepřířadí vysoký potenciál 
        # vzorkům s nulovou hustotou
        return bidder
          
            
       
    
    
    def estimate_outside(bx):
        # předpokládám, že triangulece jíž existuje
         
        # -1 = 'out', 0=success, 1=failure, 2=mix
        #current outside probability estimation
        try:
            p_out = bx.siss.estimations[-1]
        except:
            p_out = 1/(bx.nsim + 1)
            bx._logger(msg="suppose p_out=1/(bx.nsim + 1)="+str(p_out))
        
        if p_out == 0:
            p_out = 1/(bx.nsim + 1)
            bx._logger(msg="suppose p_out=1/(bx.nsim + 1)="+str(p_out))
            
        if p_out > 0.5:
            # zužovat nechceme
            # trapit sa generacema taky ne
            candidates = IS_stat.IS(bx.f, bx.h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
            
            # nevím co tam bylo za h-ko, ale nechť IM zůstane 1
            implicit_multiplicator = 1
            
        else: # tak deme... trapit sa generacema
            
            sampling_r, __ = bx.sball.get_r_iteration(p_out)
            # asi tam bylo sampling_r/bx.base_r, že?
            # u stats.norm zadáváme směrodatnou odchylku, je to asi správné
            h = f_models.UnCorD([stats.norm(0, sampling_r/bx.base_r) for i in range(bx.nvar)])
            
            # for IS_stats
            #svar = (sampling_r/bx.base_r)**2 # svar like sampling_variance
            # něco takovýho bych nahrubo placnul
            #implicit_multiplicator = svar**bx.nvar * np.exp(bx.nvar/svar - bx.nvar)
            implicit_multiplicator = np.inf
            candidates = IS_stat.IS(bx.f, h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
                
                
        bx.judge_candidates(candidates)
        # odhady zlobily, zkusím jako multiplikator posílat nekonečno
        bx.siss.add_IS_serie(candidates.w, candidates.simplex, implicit_multiplicator=implicit_multiplicator)
        
        # odebereme kus práce u selectu
        # -1 = 'out', 0=success, 1=failure, 2=mix
        candidates = candidates[candidates.event != 0] 
        candidates = candidates[candidates.event != 1] 
        # uvalíme pokutu
        bx.assess_candidates(candidates)
        # uložíme
        bx.candidates_index[-1] = candidates[candidates.event == -1] 
        if len(candidates[candidates.event == 2]) > 0:
            # -2 je určen pro zbytky, кылем-мылем
            bx.candidates_index[-2] = candidates[candidates.event == 2] 
        
    
    
    def estimate_simplex(bx, simplex_id):
        """
        Delaunay triangulation
        """
        #bx._logger(msg="estimate simplex"+str(simplex_id))
        
        simplex = bx.tri.simplices[simplex_id]
        # чылкыт f_model
        vertices = bx.f_model[simplex]
        
        
        # already divided by nsim in variance formule
        # divide by /(nvar+1)/(nvar+2) from simplex inertia tensor solution
        # multiply by simplex_volume, but it looks like it shouldn't be here
        # for simplex: d = nvar+2 
        # sice to má nazev h_plan, ale nese rozdělení a hustoty v f-ku
        h_plan = IS_stat.IS_like(vertices, sampling_space=bx.sampling_space, nis=bx.simplex_budget, d=bx.nvar+2)
        
        # nejdřív vyfiltrujeme vzorky, pak budeme řešit hustoty
        #
        h_plan_model = getattr(h_plan, bx.tri_space)
        vertices_model = getattr(vertices, bx.tri_space)
        
        # budeme pokažde sestavovat triangulaci z jedného simplexu
        # a rešit jen zda naši bodíky "inside or outside"
        # (s narustajícím nsim tohle brzy se stavá rychlejším než bežný dotaz)
        found_simplices = spatial.Delaunay(vertices_model).find_simplex(h_plan_model)
        
        # ten simplex nic neznamená, asi nebudu jej použivat
        h_plan.simplex = found_simplices
            
        # uvnitř simplexu - mel by tam bejt pouze jeden, "nulový" simplex
        mask = found_simplices == 0
        
        
        # necháme ISSI trapit sa pravděpodobnostma
        bx.siss.add_single_event_data(h_plan.w[mask], event=simplex_id, nis=bx.simplex_budget)
        
        
        # je nejvyšší čas zjistit čo to byl za utvar
        
        # fp like a failure points. Number of failure points
        # intersect 2 times faster than setdiff (in my tests)
        fp = len(np.intersect1d(simplex, bx.sample_box.failure_points, assume_unique=True))
        # -1 = 'out', 0=success, 1=failure, 2=mix
        if fp == bx.nvar + 1:
            pass
            #event = 'failure'
            #event_id = 0
        elif fp == 0:
            pass
            #event = 'success'
            #event_id = 1
        else:
            #event = 'mix'
            event_id = 2
            
            candidates = h_plan[mask]
            candidates.event = np.full(len(candidates), event_id, dtype=np.int8)
            # a vyhodnotíme je
            bx.assess_candidates(candidates)
            # vzorky je třeba přidát ke kandidatům
            # jako, nic nepokazí, ale čo tam připadně bylo - přepíše
            bx.candidates_index[simplex_id] = candidates
            
            
        
        # vzorky je třeba přidát ke kandidatům
        # zás nakladaní s odpadem...
        bx.unjudged_candidates.append(h_plan[~mask])
            
            
            
        
        
        
    
        # potřebuju nová slovesa
    def assess_candidates(bx, candidates):
        candidates.nsim_stamp = np.full(len(candidates), bx.nsim)
    
        candidates_tree = getattr(candidates, bx.tree_space)
        dd, ii = bx.tree.query(candidates_tree, k=1, p=bx.p_norm)
        candidates.dd = dd
        candidates.ii = ii
        
        PDFs = bx.sample_box.pdf(bx.tree_space)
        # teď máme hustoty kandidatů a prislušejicích jím vzorků
        PDF = PDFs[ii]
        pdf = candidates.pdf(bx.tree_space)
        
        # кучапи
        # psí-kučapí není invariántní vůči lineárním transformácím
        candidates.psee = np.sqrt(pdf*PDF) * np.power(dd , bx.nvar)
        
        # кечато
        # ksí-kěčató není invariántní vůčí rotacím
        ksee = np.empty(len(candidates))
        for i in np.unique(ii):
            # doufám, že je to legální
            ksee[i==ii] = lk.kechato_potential(bx.f_model[i], candidates[i==ii], kechato_space=bx.kechato_space)
        candidates.ksee = ksee
        
        
        
    
    
    # ještě by asi hodily funkce pro pridaní uživatelských kandidatů
    # funkce, která přídává dálší kandidaty? outside sampling
    
    
    def judge_candidates(bx, candidates):
    
        candidates_tri = getattr(candidates, bx.tri_space)
            
        found_simplices = bx.tri.find_simplex(candidates_tri)
        
        candidates.simplex = found_simplices
        events = found_simplices.copy()
        # black magic
        # zhruba - get_events() vrací pole s odpovidajícími čísly jevů pro každý simplex, počineje od nuly
        # tím slajsingem my jakoby vybirame ke každemu nalezenemu simplexovi ten správnej mu odpovídajicí jev
        events[found_simplices >= 0] = bx.simplex_events[found_simplices[found_simplices >= 0]]
        candidates.event = events
        
        
    




class OptimizedCensoredSampling(MinEnergyCensoredSampling):


    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        if bx.nsim > bx.nvar + 1:
            try:
                bx.convex_hull = spatial.ConvexHull(getattr(bx.sampled_plan, bx.tri_space), incremental=True)
            except BaseException as e:
                    msg = "error of creating ConvexHull "
                    error_msg = bx.__class__.__name__ + ": " + msg + repr(e)
                    bx._logger(msg=error_msg)
        
        super().regen()
            
            
            
    
        
        
    def increment(bx, input_sample):
        #☺ ну нахрен это ваше наследование-расследование
        
        #č nechť bude, asi až tak nevadí
        for i in range(bx.nvar):
            for j in range(len(input_sample)):
                plan_index = np.searchsorted(bx.sorted_plan_U[i], input_sample.U[j,i])
                bx.sorted_plan_U[i] = np.insert(bx.sorted_plan_U[i], plan_index, input_sample.U[j,i])
            
        #č strom posuneme sem    
        sampled_plan_tree = getattr(bx.sample_box, bx.tree_space)
        bx.tree = spatial.cKDTree(sampled_plan_tree)
        
        #č fór je v tom, že tu triangulaci nemůžeme výtvořit hned na začátku
        if "tri" in dir(bx):
            # tri - Deloneho triangulace
            
            bx.export_estimation()
            
            
            former_simplices = bx.tri.simplices
            mixed = bx.is_mixed()
            
            # sample je jíž převeden na f (v .add_sample()), takže je to bezpěčný
            bx.tri.add_points(getattr(input_sample, bx.tri_space))
            bx.simplex_events = bx.get_events()
            # print('increment se podaril')
            if len(bx.tri.coplanar): # pokud triangulace není v pořadku
                #print('triangulace v pořádku není')
                bx._logger(msg='triangulation has coplanar points')
                
            
            
            
            # zkontrolujeme co se změnilo
            # předpokladám, že se počet simplexů přidaním bodů nezměnší 
            equal_mask = former_simplices == bx.tri.simplices[:len(former_simplices)]
            changed_simplices_ids = np.argwhere(~equal_mask.all(axis=1)).reshape(-1)
            
            # invalidirujeme jejich odhady
            for simplex_id in changed_simplices_ids:
                bx.siss.delete_event_data(simplex_id)
                
                #č popajem pouze mixy, ty musel jsem spočítat před aktualizací
            for simplex_id in changed_simplices_ids[mixed[changed_simplices_ids]]:
                bx.candidates_index.pop(simplex_id)
                
            
            # pokud není splněná podmínka, 
            # tak nemáme jistotu, že se potenciály nezměni
            # ani u kandidatů, které se nacházejí v pojíštěných státem buňkách
            if (bx.tree_space != bx.tri_space) or (bx.p_norm != 2):
                for candidates in bx.candidates_index.values():
                    bx.assess_candidates(candidates)    
                
            # změněné simplexy přepočítáme
            for simplex_id in changed_simplices_ids:
                bx.estimate_simplex(simplex_id)    
                
                
                
            # teď nové simplexy
            # simplexy свежего разлива
            for simplex_id in range(len(former_simplices), bx.tri.nsimplex):
                # zde jen počítame
                bx.estimate_simplex(simplex_id)
            
            bx.convex_hull.add_points(getattr(input_sample, bx.tri_space))
            #č přepočíst -1 v zavislosti na simplexu vstupního bodu
            try:
                #č kontrola korrektní i v případě NaN
                test = input_sample.simplex > -1
                #♥ эскером
                if not test.all():
                    bx.former_candidates.append(bx.candidates_index.pop(-1))
                    bx.siss.delete_event_data(-1)
                    bx.estimate_outside()
            except BaseException as e:
                msg = "input sample didn't provide correct 'simplex' attribute "
                error_msg = bx.__class__.__name__ + ": " + msg + repr(e)
                bx._logger(msg=error_msg)
                bx.former_candidates.append(bx.candidates_index.pop(-1))
                bx.siss.delete_event_data(-1)
                bx.estimate_outside()
                
                
            
            #č prohrabeme odpad
            bx._judge_unjudged(bx.unjudged_candidates)
            bx._judge_unjudged(bx.former_candidates)
                
                
                
                
        else:
            bx._logger('Triangulace (zatím?) neexistuje')
            bx.regen()
            
    
    
    def _judge_unjudged(bx, list):
        for i in range(len(list)):
            candidates = list.pop()
            bx.is_outside(candidates)
            
            mask = candidates.is_outside
            if np.any(mask): #č pokud aspoň jeden bydlí mimo Brno
                candidates = candidates[mask]
                candidates.simplex = candidates.event = np.full(len(candidates), -1, dtype=np.int8)
                #č vyhodnotíme
                bx.assess_candidates(candidates)
                #č vzorky je třeba přidát ke kandidatům
                # jako, nic nepokazí, ale čo tam připadně bylo - přepíše
                bx.candidates_index[-1].add_sample(candidates)
    
    
    
    def estimate_outside(bx):
        # předpokládám, že triangulece jíž existuje
        
        # -1 = 'out', 0=success, 1=failure, 2=mix
        #current outside probability estimation
        try:
            p_out = bx.siss.estimations[-1]
        except:
            p_out = 1/(bx.nsim + 1)
            bx._logger(msg="suppose p_out=1/(bx.nsim + 1)="+str(p_out))
        
        if p_out == 0:
            p_out = 1/(bx.nsim + 1)
            bx._logger(msg="suppose p_out=1/(bx.nsim + 1)="+str(p_out))
        
        #
        # get candidates!
        # 
        if p_out > 0.5:
            # zužovat nechceme
            # trapit sa generacema taky ne
            candidates = IS_stat.IS(bx.f, bx.h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
            
            # nevím co tam bylo za h-ko, ale nechť IM zůstane 1
            #implicit_multiplicator = 1
            
        else: # tak jdeme... trapit sa generacema
            
            sampling_r, __ = bx.sball.get_r_iteration(p_out)
            # asi tam bylo sampling_r/bx.base_r, že?
            # u stats.norm zadáváme směrodatnou odchylku, je to asi správné
            h = f_models.UnCorD([stats.norm(0, sampling_r/bx.base_r) for i in range(bx.nvar)])
            candidates = IS_stat.IS(bx.f, h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
                
        bx.is_outside(candidates)
        candidates = candidates[candidates.is_outside]
        
        
        #č řeším problém, že při načítaní ze souboru
        #č blackbox nemá adekvatní odhad Brno-venkova
        while len(candidates) < 2:
            p_out = p_out/2
            bx._logger(msg="suppose p_out="+str(p_out))
            
            sampling_r, __ = bx.sball.get_r_iteration(p_out)
            # asi tam bylo sampling_r/bx.base_r, že?
            # u stats.norm zadáváme směrodatnou odchylku, je to asi správné
            h = f_models.UnCorD([stats.norm(0, sampling_r/bx.base_r) for i in range(bx.nvar)])
            candidates = IS_stat.IS(bx.f, h, space_from_h='R', space_to_f=bx.sampling_space,  Nsim=bx.budget)
                    
            bx.is_outside(candidates)
            candidates = candidates[candidates.is_outside]
        
        
        
        #č necháme ISSI trapit sa pravděpodobnostma
        bx.siss.add_single_event_data(candidates.w, event=-1, nis=bx.budget)
        
        candidates.simplex = candidates.event = np.full(len(candidates), -1, dtype=np.int8)
            
            
        #č vyhodnotíme
        bx.assess_candidates(candidates)
        # vzorky je třeba přidát ke kandidatům
        # jako, nic nepokazí, ale čo tam připadně bylo - přepíše
        bx.candidates_index[-1] = candidates
            
    
          

    def is_outside(bx, candidates):
    
        x = getattr(candidates, bx.tri_space)
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        A = bx.convex_hull.equations[:,:-1]
        b = bx.convex_hull.equations[:,-1]
        
        # N=nsim
        NxN = A @ x.T + np.atleast_2d(b).T
        mask = np.any(NxN > 0, axis=0)
        candidates.is_outside = mask






#č alternativní název je 
#:) UltimateFastOptimizationOptimizedMinDistanceEnergyAdaptiveCensoringSampling
class BlackSimpleX(OptimizedCensoredSampling):
    
    def increment(bx, input_sample):
        #☺ ну нахрен это ваше наследование-расследование
        
        #č nechť bude, asi až tak nevadí
        for i in range(bx.nvar):
            for j in range(len(input_sample)):
                plan_index = np.searchsorted(bx.sorted_plan_U[i], input_sample.U[j,i])
                bx.sorted_plan_U[i] = np.insert(bx.sorted_plan_U[i], plan_index, input_sample.U[j,i])
            
            
        #č strom posuneme sem    
        sampled_plan_tree = getattr(bx.sample_box, bx.tree_space)
        bx.tree = spatial.cKDTree(sampled_plan_tree)
        
        
        #č fór je v tom, že tu triangulaci nemůžeme výtvořit hned na začátku
        if "tri" in dir(bx):
            # tri - Deloneho triangulace
            
            
            
            former_simplices = bx.tri.simplices
            mixed = bx.is_mixed()
            
            # sample je jíž převeden na f (v .add_sample()), takže je to bezpěčný
            bx.tri.add_points(getattr(input_sample, bx.tri_space))
            bx.simplex_events = bx.get_events()
            # print('increment se podaril')
            if len(bx.tri.coplanar): # pokud triangulace není v pořadku
                #print('triangulace v pořádku není')
                bx._logger(msg='triangulation has coplanar points')
                
            
            
            
            # zkontrolujeme co se změnilo
            # předpokladám, že se počet simplexů přidaním bodů nezměnší 
            equal_mask = former_simplices == bx.tri.simplices[:len(former_simplices)]
            changed_simplices_ids = np.argwhere(~equal_mask.all(axis=1)).reshape(-1)
            
                
                #č popajem pouze mixy, ty musel jsem spočítat před aktualizací
            for simplex_id in changed_simplices_ids[mixed[changed_simplices_ids]]:
                bx.candidates_index.pop(simplex_id)
                
            
            # pokud není splněná podmínka, 
            # tak nemáme jistotu, že se potenciály nezměni
            # ani u kandidatů, které se nacházejí v pojíštěných státem buňkách
            if (bx.tree_space != bx.tri_space) or (bx.p_norm != 2):
                for candidates in bx.candidates_index.values():
                    bx.assess_candidates(candidates)    
                
            # změněné simplexy přepočítáme
            for simplex_id in changed_simplices_ids:
                bx.estimate_simplex(simplex_id)    
                
                
                
            # teď nové simplexy
            # simplexy свежего разлива
            for simplex_id in range(len(former_simplices), bx.tri.nsimplex):
                # zde jen počítame
                bx.estimate_simplex(simplex_id)
            
            
            bx.convex_hull.add_points(getattr(input_sample, bx.tri_space))
            #č přepočíst -1 v zavislosti na simplexu vstupního bodu
            try:
                #č kontrola korrektní i v případě NaN
                test = input_sample.simplex > -1
                #♥ эскером
                if not test.all():
                    bx.former_candidates.append(bx.candidates_index.pop(-1))
                    bx.siss.delete_event_data(-1)
                    bx.estimate_outside()
            except BaseException as e:
                msg = "input sample didn't provide correct 'simplex' attribute "
                error_msg = bx.__class__.__name__ + ": " + msg + repr(e)
                bx._logger(msg=error_msg)
                bx.former_candidates.append(bx.candidates_index.pop(-1))
                bx.siss.delete_event_data(-1)
                bx.estimate_outside()
                
                
            
            #č prohrabeme odpad
            bx._judge_unjudged(bx.unjudged_candidates)
            bx._judge_unjudged(bx.former_candidates)
                
            
            
            #č A ještě...
            # 
            highest_bid = 0 
            for i, candidates in bx.candidates_index.items():
                bids = getattr(candidates, bx.potencial)
                bid = np.max(bids)
                # side effect
                if bid > highest_bid:
                    if np.any(candidates.nsim != bx.nsim):
                        if i == -1:
                            bx.assess_candidates(candidates)  
                        else:
                            bx.estimate_simplex(i)
                            candidates = bx.candidates_index[i]
                            
                        #š a eště ráz
                        bids = getattr(candidates, bx.potencial)
                        bid = np.max(bids)
                        if bid > highest_bid:
                            highest_bid = bid
                    
                    else:
                        highest_bid = bid
                
                
        else:
            bx._logger('Triangulace (zatím?) neexistuje')
            bx.regen()
            
            
            
        
    #č pro kompatibilitu s bx.regen() rodičovské třidy
    #č nechávám stejné názvy, ale žádný estimate tu nedělám
    def estimate_simplex(bx, simplex_id):
        
        simplex = bx.tri.simplices[simplex_id]
        
        # fp like a failure points. Number of failure points
        # intersect 2 times faster than setdiff (in my tests)
        fp = len(np.intersect1d(simplex, bx.samplebox.failure_points, assume_unique=True))
        #č hned na začátku odmítneme simplex, 
        #č pokud nevypadá jako stará Běloruská vlajka
        if (fp > 0) and (fp < bx.nvar + 1):
            
            # чылкыт f_model
            vertices = bx.f_model[simplex]
            
            
            vertices_model = getattr(vertices, bx.tri_space)
            
            #č budeme pokažde sestavovat ConvexHull z jedného simplexu
            #č a rešit jen zda naši bodíky "inside or outside"
            #č (s narustajícím nsim tohle brzy se stavá rychlejším než bežný dotaz)
            convex_hull = spatial.ConvexHull(vertices_model)
            
            
            #č nepodařilo se mi ve FORTRANovém kódě Kobyly význat #ё(древние руны, блин)
            #č a tak nevím jak přesně Kobyla sestavuje simplex.
            #č Každopádně, jsme asi schopní nabídnout lepší "roběh",
            #č než s tou implicitnou jedničkou.
            x0 = np.mean(vertices_model, axis=0)
            rhobeg = np.mean(np.abs(vertices_model - x0))
            #print("roběh", rhobeg)
            # catol - Tolerance (absolute) for constraint violations
            # default value for catol is 0.0002
            options = {'rhobeg': rhobeg, 'maxiter': bx.simplex_budget, 'disp': False, 'catol': 0}
            
            
            constraints = six.get_COBYLA_constraints(convex_hull)
            # as I can see, scipy treats 'tol' as 'rhoend'
            # with default value tol=1e-4
            #č jediný mechnizmus ukončení - vyčerpaní maximálního počtu iterací
            res = optimize.minimize(bx.assess_candidate, x0, args=(), method='COBYLA', \
                        tol=0, constraints=constraints, options=options)
                        
            bx._logger("simplex %s optimization: %s" %(simplex_id, res))
            #bx._logger(msg="estimate simplex"+str(simplex_id))
        
            
            x_sample = bx.f_model.new_sample(res.x, bx.tri_space)
            candidate = CandyBox(x_sample, event=[2])
            #č a, kruci, je třeba zas je vyhodnotit
            bx.assess_candidates(candidate)
            # vzorky je třeba přidát ke kandidatům
            # jako, nic nepokazí, ale čo tam připadně bylo - přepíše
            bx.candidates_index[simplex_id] = candidate
            

    
    
    def assess_candidate(bx, x):
        """
        Objective function for optimalization
        """
        
        if (bx.tri_space == bx.tree_space) and (bx.potencial == 'psee'):
            x_tree = x
            pdf = float(bx.f_model.sample_pdf(x, bx.tree_space))
        else:#č máme smulu
            x_sample = bx.f_model.new_sample(x, bx.tri_space)
            x_tree = getattr(x_sample, bx.tree_space)
            pdf = float(x_sample.pdf(bx.tree_space))
        
        #č dd a ii budou prostě skalární hodnoty
        dd, ii = bx.tree.query(x_tree, k=1, p=bx.p_norm)
        
        PDFs = bx.sample_box.pdf(bx.tree_space)
        # teď máme hustoty kandidatů a prislušejicích jím vzorků
        PDF = PDFs[ii]
        
        if bx.potencial == 'psee':
            #оӵ кучапи
            #č pejskovej potenciál
            #č psí-kučapí není invariántní vůči lineárním transformácím
            
            psee = -np.sqrt(pdf*PDF) * np.power(dd , bx.nvar)
            #print(psee)
            return psee
        else: # ksee 
            #оӵ кечато
            #č koťatko-káčátkovej potenciál
            #č ksí-kěčató není invariántní vůčí rotacím
            
            ksee = -float(lk.kechato_potential(bx.f_model[ii], x_sample, kechato_space=bx.kechato_space)) 
            #print(ksee)
            return ksee



#♥♥♥♥♥♥
# BlackBoxiCheck
class KechatoLukiskon(BlackBox):
    def __init__(kl, sample_object, model_space='Rn', sampling_space=None, kechato_space='Rn', p_norm=1, gradient=None, budget=20000):
        kl.model_space = model_space
        kl.sampling_space = sampling_space
        kl.kechato_space = kechato_space
        kl.p_norm = p_norm
        kl.gradient = gradient
        kl.budget = budget
        
        super().__init__(sample_object)
        
        
    def __call__(kl):
        kl._logger(msg="умой! Умойлэсь но умой сэмпл шедьтоно")
        
        if (len(kl.failure_points) < 1) or (len(kl.success_points) < 1):
            kl._logger(msg="Умоез сэмпл шедьтыны уг быгаты :( ")
            return kl.LHS_like_correction(kl.h(1))
            
        else:
            
            kl.ivortodon = 0
            kl.to_sample = None
            
            # ty brdo, stm nedává číslo tečky. Nemusí, no...
            kl.i = 0
            stm.Voronoi_2_point_estimation(kl, model_space=kl.model_space, sampling_space=kl.sampling_space,\
                                        p_norm=kl.p_norm, gradient=kl.gradient, budget=kl.budget, callback=kl.callback)
                
            if kl.to_sample is not None:
                # hrubě ošidíme s takovejhlema usilima námi pěčlivě vybraný vzorečiček
                return kl.LHS_like_correction(kl.to_sample) 
            else:
                kl._logger(msg="Умоез сэмпл шедьтыны ӧй быгаты :( ")
                return kl.LHS_like_correction(kl.h(1))
        
        
    def callback(kl, *args, **kwargs):
        # spoléhám na to, že stm vrácí tečky v f-ku
        nodes = kwargs['nodes']
        node_pf = nodes.node_pf_estimations
        
        mask = (node_pf > 0) * (node_pf < 1)
        nodes = nodes[mask]
        if len(nodes) > 0:
            node_pf = node_pf[mask]
            
            #  !!!! ENTROPY !!!
            nodes.entropy = -node_pf*np.log(node_pf) - (1-node_pf)*np.log(1-node_pf)
            
            nodes.ksee = lk.kechato_potential(kl.f_model[kl.i], nodes, kechato_space=kl.kechato_space)
            
            nodes.ivortodon = nodes.entropy * nodes.ksee
            
            if np.max(nodes.ivortodon) > kl.ivortodon:
                # side effects
                kl.ivortodon = np.max(nodes.ivortodon)
                kl.to_sample = nodes[np.argmax(nodes.ivortodon)]
        
        # side effect
        kl.i += 1
                                    
    
#♥♥♥♥♥♥♥♥♥♥♥♥
class KechatoTwoPointLukiskon(BlackBox):
    def __init__(kl, sample_object, model_space='Rn', sampling_space=None, kechato_space='Rn', p_norm=1, gradient=None, budget=20000):
        kl.model_space = model_space
        kl.sampling_space = sampling_space
        kl.kechato_space = kechato_space
        kl.p_norm = p_norm
        kl.gradient = gradient
        kl.budget = budget
        
        super().__init__(sample_object)
        
        
    def __call__(kl):
        kl._logger(msg="умой! Умойлэсь но умой сэмпл шедьтоно")
        
        kl.candidates_index = {}
        
        if (len(kl.failure_points) < 1) or (len(kl.success_points) < 1):
            kl._logger(msg="Умоез сэмпл шедьтыны уг быгаты :( ")
            return kl.LHS_like_correction(kl.h(1))
            
        else:
            
            kl.ivortodon = 0
            kl.to_sample = None
            
            # ty brdo, stm nedává číslo tečky. Nemusí, no...
            kl.i = 0
            stm.Voronoi_2_point_estimation(kl, model_space=kl.model_space, sampling_space=kl.sampling_space,\
                                        p_norm=kl.p_norm, gradient=kl.gradient, budget=kl.budget, callback=kl.callback)
                
            if kl.to_sample is not None:
                # hrubě ošidíme s takovejhlema usilima námi pěčlivě vybraný vzorečiček
                return kl.LHS_like_correction(kl.to_sample) 
            else:
                kl._logger(msg="Умоез сэмпл шедьтыны ӧй быгаты :( ")
                return kl.LHS_like_correction(kl.h(1))
        
        
    def callback(kl, *args, **kwargs):
        # spoléhám na to, že stm vrácí tečky v f-ku
        nodes = kwargs['nodes']
        node_pf = nodes.node_pf_estimations
        
        mask = (node_pf > 0) * (node_pf < 1)
        nodes = nodes[mask]
        if len(nodes) > 0:
            node_pf = node_pf[mask]
            
            #  !!!! ENTROPY !!!
            nodes.entropy = -node_pf*np.log(node_pf) - (1-node_pf)*np.log(1-node_pf)
            
            
            #nodes.ksee = lk.kechato_potential(kl.f_model[kl.i], nodes, kechato_space=kl.kechato_space)
            #♥ кечато
            #č ksí-kěčató není invariántní vůčí rotacím
            ksee = np.empty(len(nodes))
            ii2 = nodes.ii2
            for ii in np.unique(nodes.ii2):
                #č doufám, že je to legální
                ksee[ii==ii2] = lk.kechato_potential(kl.f_model[[kl.i, ii]], nodes[ii==ii2], kechato_space=kl.kechato_space)
            nodes.ksee = ksee
            
            
            nodes.ivortodon = nodes.entropy * nodes.ksee
            
            kl.candidates_index[kl.i] = nodes
            
            if np.max(nodes.ivortodon) > kl.ivortodon:
                #E side effects
                kl.ivortodon = np.max(nodes.ivortodon)
                kl.to_sample = nodes[np.argmax(nodes.ivortodon)]
        
        #E side effect
        kl.i += 1    






#!/usr/bin/env python
# coding: utf-8


"""
 cs: 
 

 en: 

"""

import numpy as np
from scipy import stats
import copy

class WellMetError(ValueError):
    pass

# alpha atribut beztak veřejný, ale tato funkce ještě provadí normalizaci a kontrolu
def set_alpha(f_model, input_alpha):
    alpha = np.atleast_1d(input_alpha).reshape(-1)
    if len(alpha) != f_model.nvar:
        raise WellMetError
    else:
        f_model.alpha = alpha / stats.gmean(alpha)

class Ingot:
    """
    Prazdná třida pro "nevypalené" vzorky, tj. bez přiřazeného rozdělení
    """
    def __init__(self, data, attr='R'):
        self.attr = attr
        try:
            self.data = np.atleast_2d(getattr(data, attr))
        except AttributeError:
            self.data = np.atleast_2d(data)
        
        
    def __repr__(self):
        return "%s(np.%s, '%s')"%('Ingot', repr(self.data), self.attr)
        
    def __str__(self):
        return str(self.data)
        
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, slice):
        self_copy = copy.copy(self)
        # robim kopiu z _data, nebo ne?
        self_copy.data = np.atleast_2d(self.data[slice,:]) #.copy()
        return self_copy
        
        
    def __getattr__(self, attr):
        if attr == 'f_model':
            return self
        # Рекурсилы пезьдэт!
        elif attr in ('attr', 'data'):
            raise AttributeError(attr)
        elif attr == 'nvar':
            nsim, nvar = self.data.shape
            return nvar
        elif attr == 'nsim':
            return len(self.data)
            
        # гулять так гулять
        elif attr == self.attr:
            return self.data
            
        raise AttributeError(attr)
      
        
        # vstupné vzorky jsou implicitně R, 
        # nikoliv z prostoru tohoto krámu
    def add_sample(self, sample, space='R'):
        # first of all, are you one of us?
        if space == self.attr:
            # does sample is another sample object? 
            # zde nechcu, aby spadlo
            try:
                self.data = np.vstack((self.data, getattr(sample, self.attr)))
            except AttributeError:
                self.data = np.vstack((self.data, sample))
            
        # no, actually
        else:
            # does sample is another sample object? 
            # self.data = np.vstack((self.data, getattr(sample, self.attr)))
            # ale zde chcu 
            # (aby spadlo)
            raise WellMetError
            
        
    # drobná pomucka
    def new_sample(f, sample=None, space='R'):  
        return Ingot(sample, space) # too easy








class SNorm:
    __slots__ = ('G', '_U', '_pdf_G', 'alpha')
    """
    Standard Gauss distribution
    """
    @classmethod
    def _fromdata(cls, G, U=None, pdf_G=None, alpha=1):
        f = super().__new__(cls)
        f.G = G
        f._U = U
        f._pdf_G = pdf_G
        f.alpha = alpha
        return f
    
    @classmethod
    def from_G(cls, G, alpha=1):
        f = super().__new__(cls)
        f.G = G
        f._U = None
        f._pdf_G = None
        f.alpha = alpha
        return f
        
    @classmethod
    def from_U(cls, U, alpha=1):
        f = super().__new__(cls)
        f._U = U
        f.G = stats.norm.ppf(U)
        f._pdf_G = None
        f.alpha = alpha
        return f
    
    def __init__(self, nvar, alpha=1):
        self.G = np.empty((0, nvar), dtype=float)
        self._U = None
        self._pdf_G = None
        self.alpha = alpha
        
     # nemusím duplikovat a dědit 
    set_alpha = set_alpha
    
    def __repr__(self):
        return "%s(%s, %s)"%('SNorm', self.nvar, repr(self.alpha))
        
    def __str__(f):
        return "SNorm sample: " + str(f.R)
        
    def __call__(self, ns=0):
        if ns:
            return SNorm.from_G(np.random.randn(ns, self.nvar), self.alpha)
        #else:
        return SNorm(self.nvar, self.alpha)
        
    def __len__(self):
        return len(self.G)
    
    def __getitem__(self, slice):
        G = np.atleast_2d(self.G[slice]) 
        
        if self._U is not None:
            U = np.atleast_2d(self._get_U()[slice]) 
        else:
            U = None
                
        if self._pdf_G is not None:
            pdf_G = np.atleast_1d(self._get_pdf()[slice]) 
        else:
            pdf_G = None
                
        return SNorm._fromdata(G, U, pdf_G, self.alpha)
        
        
    def __getattr__(f, attr):
        if attr == 'f_model':
            return f
        elif attr in ('R', 'Rn', 'GK'):
            return f.G
        elif attr in ('aR', 'aRn', 'aGK', 'aG'):
            return f.G * f.alpha
        elif attr in ('P', 'U'):
            return f._get_U()
        elif attr in ('aP', 'aU'):
            return f.U * f.alpha
        elif attr == 'nvar':
            return f.G.shape[1]
        elif attr == 'nsim':
            return len(f.G)
        elif attr == 'marginals':
            return [stats.norm for __ in range(f.nvar)]
        elif attr == 'cor':
            return np.diag([1 for __ in range(f.nvar)])
            
        # hustoty
        # I'm considering to deprecate attribute access
        elif attr[:4] == 'pdf_':
            return f.pdf(attr[4:])
            
        raise AttributeError(attr)
        
    
    def _get_U(self):
        _U = self._U
        G = self.G
        if _U is None:
            U = stats.norm.cdf(G)
            self._U = U
            return U
            
        len_U = len(_U)
        if len_U == len(G):
            return _U
        else:
            U = np.empty_like(G)
            U[:len_U] = _U
            U[len_U:] = stats.norm.cdf(G[len_U:])
            self._U = U
            return U
            
            
    def _get_pdf(self):
        _pdf_G = self._pdf_G
        G = self.G
        if _pdf_G is None:
            pdf_G = self._calculate_pdf(G)
            self._pdf_G = pdf_G
            return pdf_G
            
        len_pdf = len(_pdf_G)
        nsim = len(G)
        if len_pdf == nsim:
            return _pdf_G
        else:
            pdf_G = np.empty(nsim, dtype=float)
            pdf_G[:len_pdf] = _pdf_G
            pdf_G[len_pdf:] = self._calculate_pdf(G[len_pdf:])
            self._pdf_G = pdf_G
            return pdf_G
            
            
    @staticmethod           
    def _calculate_pdf(G):
        return np.prod(stats.norm.pdf(G), axis=1)
            
            
    def _add_U(self, U_to_add):
        "č method přeppokladá, že self.G ještě není dotčen"
        n = len(U_to_add)
        nsim = self.nsim
        U = np.empty((nsim + n, self.nvar), dtype=float)
        U[nsim:] = U_to_add
        
        _U = self._U
        G = self.G
        if _U is None:
            U[:nsim] = stats.norm.cdf(G)
        else:
            len_U = len(_U)
            if len_U == len(G):
                U[:nsim] = _U
            else:
                U[:len_U] = _U
                U[len_U:nsim] = stats.norm.cdf(G[len_U:])
            
        self._U = U
            
        
    def add_sample(f, input_sample, space='R'):
        """
        Adds coordinates from input sample
        """
        try: # is sample another f_model object? 
            sample = getattr(input_sample, space)
        except AttributeError:
            # no, it is just coordinates array
            sample = input_sample
        
        sample = np.atleast_2d(sample)
        isim, ivar = np.shape(sample) # input sim, var
        if ivar != f.nvar:
            raise WellMetError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
        
        if isim == 0:
            return None
        
        nsim = f.nsim
        G = np.empty((nsim + isim, f.nvar), dtype=float)
        G[:nsim] = f.G
        
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample = sample / f.alpha
            space = space[1:]
        
        if space in ('R', 'Rn', 'GK', 'G'):
            G[nsim:] = sample
            f.G = G
        elif space in ('P', 'U'):
            #č _add_U() přeppokladá, že self.G ještě není dotčen
            f._add_U(sample)
            G[nsim:] = stats.norm.ppf(sample)
            f.G = G
        else:
            raise WellMetError('SNorm: unknown space %s' % space)
            
        
    #č tohle už není "drobná pomucka"
    #č to je jedná z funkcí, která běží 30% času
    def new_sample(f, input_sample=None, space='R', extend=False):
        """
        Returns new f_model object with the same distribution and with coordinates from 'sample' taken
        """
        if input_sample is None:
            return SNorm(f.nvar, f.alpha)
        
        try: # does sample is another f_model object? 
            sample = getattr(input_sample, space)
        except AttributeError:
            # no, it is just coordinates array
            sample = input_sample
        
        sample = np.atleast_2d(sample)
        isim, ivar = np.shape(sample) # input sim, var
        if extend:
            to_extend = sample
            sample = np.zeros((isim, f.nvar))
            sample[:,:ivar] = to_extend
        elif ivar != f.nvar:
            raise WellMetError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
            
        
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample = sample / f.alpha
            space = space[1:]
        
        
        if space in ('R', 'Rn', 'GK', 'G'):
            return SNorm.from_G(sample, f.alpha)
        elif space in ('P', 'U'):
            return SNorm.from_U(sample, f.alpha)
        else:
            raise WellMetError('SNorm: unknown space %s' % space)
            
    
    
    
    
    def pdf(f, space='R'):
        """
        Returns own probability densities in the given space
        """
        if space in ('R', 'Rn', 'GK', 'G'):
            return f._get_pdf()
        elif space in ('P', 'U'):
            ones = np.ones(f.nsim)
            mask = np.all((f.U >= 0) & (f.U <= 1), axis=1)
            if not np.all(mask):
                ones[~mask] = 0
            return ones
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.pdf(space[1:]) / np.prod(f.alpha)
            
        else:   
            raise WellMetError('Unknown space %s' % space)
            
        
        
    def sample_pdf(f, input_sample, space='R'):
        """
        Calculates probability density for the given external sample in the given space.
        Function intended for the case no transformation needed. 
        Otherwise new f_sample should be instanciated.
        """
        
        # does sample is another f_model object? 
        # кинлы со zase кулэ?
        try:
            sample = getattr(input_sample, space)
        except AttributeError:
            # no
            sample = np.atleast_2d(input_sample)
            
        
        if space in ('R', 'Rn', 'GK', 'G'):
            pdfs_R = stats.norm.pdf(sample)
            return np.prod(pdfs_R, axis=pdfs_R.ndim-1)
            
        elif space in ('P', 'U'):
            ones = np.ones(len(sample))
            mask = np.all((sample >= 0) & (sample <= 1), axis=1)
            if not np.all(mask):
                ones[~mask] = 0
            return ones
            
        # new piece of code "alpha"-related
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.sample_pdf(sample / f.alpha, space=space[1:]) / np.prod(f.alpha)
            
        else:   
            raise WellMetError('Unknown space %s' % space)



class Norm:
    __slots__ = ('G', '_U', '_pdf_G', 'mean', 'std', 'alpha')
    """
    Normal distribution
    """
    @classmethod
    def _fromdata(cls, G, U, pdf_G, mean, std, alpha=1):
        f = super().__new__(cls)
        f.G = G
        f._U = U
        f._pdf_G = pdf_G
        f.mean = mean
        f.std = std
        f.alpha = alpha
        return f
    
    @classmethod
    def from_G(cls, G, mean, std, alpha=1):
        f = super().__new__(cls)
        f.G = G
        f.mean = mean
        f.std = std
        f._U = None
        f._pdf_G = None
        f.alpha = alpha
        return f
        
    @classmethod
    def from_U(cls, U, mean, std, alpha=1):
        f = super().__new__(cls)
        f._U = U
        f.G = stats.norm.ppf(U)
        f.mean = mean
        f.std = std
        f._pdf_G = None
        f.alpha = alpha
        return f
    
    def __init__(self, mean, std, alpha=1):
        assert len(mean) == len(std)
        self.mean = mean
        self.std = std
        
        self.G = np.empty((0, len(mean)), dtype=float)
        self._U = None
        self._pdf_G = None
        self.alpha = alpha
        
     # nemusím duplikovat a dědit 
    set_alpha = set_alpha
    
    def __repr__(self):
        return "%s(%s, %s, %s)"%('Norm', self.mean, self.std, repr(self.alpha))
        
    def __str__(f):
        return "Norm sample: " + str(f.R)
        
    def __call__(self, ns=0):
        if ns:
            return Norm.from_G(np.random.randn(ns, self.nvar), self.mean, self.std, self.alpha)
        #else:
        return Norm(self.mean, self.std, self.alpha)
        
    def __len__(self):
        return len(self.G)
    
    def __getitem__(self, slice):
        G = np.atleast_2d(self.G[slice]) 
        
        if self._U is not None:
            U = np.atleast_2d(self._get_U()[slice]) 
        else:
            U = None
                
        if self._pdf_G is not None:
            pdf_G = np.atleast_1d(self._get_pdf()[slice]) 
        else:
            pdf_G = None
                
        return Norm._fromdata(G, U, pdf_G, self.mean, self.std, self.alpha)
        
        
    def __getattr__(f, attr):
        if attr == 'f_model':
            return f
        elif attr == 'R':
            sample = f.G * f.std
            sample += f.mean
            return sample
        elif attr in ('Rn', 'GK'):
            return f.G
        elif attr in ('aR', 'aRn', 'aGK', 'aG'):
            return f.G * f.alpha
        elif attr in ('P', 'U'):
            return f._get_U()
        elif attr in ('aP', 'aU'):
            return f.U * f.alpha
        elif attr == 'nvar':
            return len(f.mean)
        elif attr == 'nsim':
            return len(f.G)
        elif attr == 'marginals':
            return [stats.norm(m, s) for m, s in zip(self.mean, self.std)]
        elif attr == 'cor':
            return np.diag([1 for __ in range(f.nvar)])
            
        # hustoty
        # I'm considering to deprecate attribute access
        elif attr[:4] == 'pdf_':
            return f.pdf(attr[4:])
            
        raise AttributeError(attr)
        
    
    def _get_U(self):
        _U = self._U
        G = self.G
        if _U is None:
            U = stats.norm.cdf(G)
            self._U = U
            return U
            
        len_U = len(_U)
        if len_U == len(G):
            return _U
        else:
            U = np.empty_like(G)
            U[:len_U] = _U
            U[len_U:] = stats.norm.cdf(G[len_U:])
            self._U = U
            return U
            
            
    def _get_pdf(self):
        _pdf_G = self._pdf_G
        G = self.G
        if _pdf_G is None:
            pdf_G = self._calculate_pdf(G)
            self._pdf_G = pdf_G
            return pdf_G
            
        len_pdf = len(_pdf_G)
        nsim = len(G)
        if len_pdf == nsim:
            return _pdf_G
        else:
            pdf_G = np.empty(nsim, dtype=float)
            pdf_G[:len_pdf] = _pdf_G
            pdf_G[len_pdf:] = self._calculate_pdf(G[len_pdf:])
            self._pdf_G = pdf_G
            return pdf_G
            
            
    @staticmethod           
    def _calculate_pdf(G):
        return np.prod(stats.norm.pdf(G), axis=1)
            
            
    def _add_U(self, U_to_add):
        "č method přeppokladá, že self.G ještě není dotčen"
        n = len(U_to_add)
        nsim = self.nsim
        U = np.empty((nsim + n, self.nvar), dtype=float)
        U[nsim:] = U_to_add
        
        _U = self._U
        G = self.G
        if _U is None:
            U[:nsim] = stats.norm.cdf(G)
        else:
            len_U = len(_U)
            if len_U == len(G):
                U[:nsim] = _U
            else:
                U[:len_U] = _U
                U[len_U:nsim] = stats.norm.cdf(G[len_U:])
            
        self._U = U
            
        
    def add_sample(f, input_sample, space='R'):
        """
        Adds coordinates from input sample
        """
        try: # is sample another f_model object? 
            sample = getattr(input_sample, space)
        except AttributeError:
            # no, it is just coordinates array
            sample = input_sample
        
        sample = np.atleast_2d(sample)
        isim, ivar = np.shape(sample) # input sim, var
        if ivar != f.nvar:
            raise WellMetError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
        
        if isim == 0:
            return None
        
        nsim = f.nsim
        G = np.empty((nsim + isim, f.nvar), dtype=float)
        G[:nsim] = f.G
        
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample = sample / f.alpha
            space = space[1:]
        
        if space in ('Rn', 'GK', 'G'):
            G[nsim:] = sample
            f.G = G
        elif space == 'R':
            G[nsim:] = sample
            G[nsim:] -= f.mean
            G[nsim:] /= f.std
            f.G = G
        elif space in ('P', 'U'):
            #č _add_U() přeppokladá, že self.G ještě není dotčen
            f._add_U(sample)
            G[nsim:] = stats.norm.ppf(sample)
            f.G = G
        else:
            raise WellMetError('SNorm: unknown space %s' % space)
            
        
    #č tohle už není "drobná pomucka"
    #č to je jedná z funkcí, která běží 30% času
    def new_sample(f, input_sample=None, space='R', extend=False):
        """
        Returns new f_model object with the same distribution and with coordinates from 'sample' taken
        """
        if input_sample is None:
            return Norm(f.mean, f.std, f.alpha)
        
        try: # does sample is another f_model object? 
            sample = getattr(input_sample, space)
        except AttributeError:
            # no, it is just coordinates array
            sample = input_sample
        
        sample = np.atleast_2d(sample)
        isim, ivar = np.shape(sample) # input sim, var
        if extend:
            to_extend = sample
            sample = np.zeros((isim, f.nvar))
            sample[:,:ivar] = to_extend
        elif ivar != f.nvar:
            raise WellMetError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
            
        
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample = sample / f.alpha
            space = space[1:]
        
        
        if space in ('Rn', 'GK', 'G'):
            return Norm.from_G(sample, f.mean, f.std, f.alpha)
        elif space == 'R':
            sample = sample - f.mean
            sample /= f.std
            return Norm.from_G(sample, f.mean, f.std, f.alpha)
        elif space in ('P', 'U'):
            return Norm.from_U(sample, f.mean, f.std, f.alpha)
        else:
            raise WellMetError('SNorm: unknown space %s' % space)
            
    
    
    
    
    def pdf(f, space='R'):
        """
        Returns own probability densities in the given space
        """
        if space in ('Rn', 'GK', 'G'):
            return f._get_pdf()
        elif space == 'R':
            return f._get_pdf() / np.prod(f.std)
        elif space in ('P', 'U'):
            ones = np.ones(f.nsim)
            mask = np.all((f.U >= 0) & (f.U <= 1), axis=1)
            if not np.all(mask):
                ones[~mask] = 0
            return ones
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.pdf(space[1:]) / np.prod(f.alpha)
            
        else:   
            raise WellMetError('Unknown space %s' % space)
            
        
        
    def sample_pdf(f, input_sample, space='R'):
        """
        Calculates probability density for the given external sample in the given space.
        Function intended for the case no transformation needed. 
        Otherwise new f_sample should be instanciated.
        """
        
        # does sample is another f_model object? 
        # кинлы со zase кулэ?
        try:
            sample = getattr(input_sample, space)
        except AttributeError:
            # no
            sample = np.atleast_2d(input_sample)
            
        
        if space in ('Rn', 'GK', 'G'):
            pdfs_R = stats.norm.pdf(sample)
            return np.prod(pdfs_R, axis=pdfs_R.ndim-1)
            
        elif space == 'R':
            sample = sample - f.mean
            sample /= f.std
            pdfs_R = stats.norm.pdf(sample)
            return np.prod(pdfs_R, axis=pdfs_R.ndim-1) / np.prod(f.std)
            
        elif space in ('P', 'U'):
            ones = np.ones(len(sample))
            mask = np.all((sample >= 0) & (sample <= 1), axis=1)
            if not np.all(mask):
                ones[~mask] = 0
            return ones
            
        # new piece of code "alpha"-related
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.sample_pdf(sample / f.alpha, space=space[1:]) / np.prod(f.alpha)
            
        else:   
            raise WellMetError('Unknown space %s' % space)





class SNormClassic:
    """
    Standard Gauss distribution
    """
    sample_pdf = SNorm.sample_pdf
    
    def __init__(self, nvar, alpha=None):
        self.__nvar = nvar
        # nvar_R + nvar_U + pdf_R
        rowsize = nvar*2 + 1
        # data? takový neslaný nazev...
        # data suppose to be cKDTree compatible, i.e. 
        # nsim, nvar = data.shape
        self._data = np.empty((0, rowsize), dtype=float)
        if alpha is None:
            self.alpha = np.ones(nvar)
        else:
            self.set_alpha(alpha)
        
     # nemusím duplikovat a dědit 
    set_alpha = set_alpha
    
    def __repr__(self):
        return "%s(%s, %s)"%('SNorm', self.__nvar, repr(self.alpha))
        
    def __str__(f):
        return "SNorm sample: " + str(f.R)
        
    def __call__(f, ns=0):
        f_copy = copy.copy(f)
        if ns:
            sample_P = np.random.random((ns, f_copy.nvar))
            sample_R = stats.norm.ppf(sample_P)
                
            pdfs_R = stats.norm.pdf(sample_R)
            pdf_R = np.prod(pdfs_R, axis=1).reshape(-1, 1)
            f_copy._data = np.hstack((sample_R, sample_P, pdf_R))
        else:
            f_copy._data = np.empty((0, f_copy._data.shape[1]), dtype=float)
        return f_copy
        
    def __len__(f):
        return len(f._data)
    
    def __getitem__(f, slice):
        f_copy = copy.copy(f)
        # robim kopiu z _data, nebo ne?
        f_copy._data = np.atleast_2d(f._data[slice,:]) #.copy()
        return f_copy
        
        
    def __getattr__(f, attr):
        if attr == 'f_model':
            return f
        elif attr in ('R', 'Rn', 'GK', 'G'):
            return f._data[:,:f.__nvar]
        elif attr in ('aR', 'aRn', 'aGK', 'aG'):
            return f._data[:,:f.__nvar] * f.alpha
        elif attr in ('P', 'U'):
            return f._data[:,f.__nvar:2*f.__nvar]
        elif attr in ('aP', 'aU'):
            return f._data[:,f.__nvar:2*f.__nvar] * f.alpha
        elif attr == 'nvar':
            return f.__nvar
        elif attr == 'nsim':
            return len(f._data)
        elif attr == 'marginals':
            return [stats.norm for __ in range(f.__nvar)]
        elif attr == 'cor':
            return np.diag([1 for __ in range(f.__nvar)])
            
        # hustoty
        # I'm considering to deprecate attribute access
        elif attr[:4] == 'pdf_':
            return f.pdf(attr[4:])
            
        raise AttributeError(attr)
        
        
       # pro určitou konzistenci. Ne že bych chtěl zamykat dveře v Pythonu
    def __setattr__(f, attr, value):
        if attr in ('_SNorm__nvar','_data', 'alpha'):
            f.__dict__[attr] = value
        else:
            #raise AttributeError('Čo tu robíš?')
            #raise AttributeError('Враг не пройдёт!')
            #raise AttributeError('Иди отсюда!')
            #raise AttributeError('Аслыкъёсы воштыны уг луи!')
            raise AttributeError('Atribute %s of %s object is not writable' % (attr, f.__class__.__name__))
        
        
        
    def add_sample(f, sample, space='R'):
        """
        Adds coordinates from input sample
        """
        newdata = f._parse_sample(sample, space)
        f._data = np.vstack((f._data, newdata))
            
        
    #č tohle už není "drobná pomucka"
    #č to je jedná z funkcí, která běží 30% času
    def new_sample(f, sample=None, space='R', extend=False):
        """
        Returns new f_model object with the same distribution and with coordinates from 'sample' taken
        """
        f_copy = copy.copy(f)
        if sample is None:
            f_copy._data = np.empty((0, f_copy._data.shape[1]), dtype=float)
        else:
            f_copy._data = np.atleast_2d(f._parse_sample(sample, space, extend))
        return f_copy
    
    
    def _parse_sample(f, input_sample, space='R', extend=False):
        # does sample is exactly me?
        if f.__repr__() == input_sample.__repr__():
            newdata = input_sample._data
            
        else:
            try: # does sample is another f_model object? 
                sample = getattr(input_sample, space)
            except AttributeError:
                # no, it is just coordinates array
                sample = input_sample
            
            sample = np.atleast_2d(sample)
            isim, ivar = np.shape(sample) # input sim, var
            if extend:
                to_extend = sample
                sample = np.zeros((isim, f.nvar))
                sample[:,:ivar] = to_extend
            elif ivar != f.nvar:
                raise ValueError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
                
            
            # new piece of code "alpha"-related
            if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
                sample = sample / f.alpha
                space = space[1:]
            
            if space in ('R', 'Rn', 'GK', 'G'):
                sample_R = sample
                sample_P = stats.norm.cdf(sample_R)
                
                pdfs_R = stats.norm.pdf(sample_R)
                pdf_R = np.prod(pdfs_R, axis=pdfs_R.ndim-1)
                if pdfs_R.ndim == 2:
                    newdata = np.hstack((sample_R, sample_P, pdf_R.reshape(len(pdf_R), 1)))
                else:
                    newdata = np.hstack((sample_R, sample_P, pdf_R))
                
            elif space in ('P', 'U'):
                sample_P = sample
                sample_R = stats.norm.ppf(sample_P)
                    
                pdfs_R = stats.norm.pdf(sample_R)
                pdf_R = np.prod(pdfs_R, axis=pdfs_R.ndim-1)
                if pdfs_R.ndim == 2:
                    newdata = np.hstack((sample_R, sample_P, pdf_R.reshape(len(pdf_R), 1)))
                else:
                    newdata = np.hstack((sample_R, sample_P, pdf_R))
                    
            else:
                raise ValueError('SNorm: unknown space %s' % space)
            
        return newdata
    
    
    
    
    def pdf(f, space='R'):
        """
        Returns own probability densities in the given space
        """
        if space in ('R', 'Rn', 'GK', 'G'):
            return f._data[:,-1]
        elif space in ('P', 'U'):
            return f.sample_pdf(f, space)
            
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.pdf(space[1:]) / np.prod(f.alpha)
            
        else:   
            raise ValueError('Unknown space %s' % space)
            
        
    
        
            
    


class UnCorD: # nic moc nazev, ale je přece lepší nez CommonJointDistribution
    """
    Takes tuple of scipy stats distribution objects
    """
    def __init__(self, marginals, alpha=None):
        self.__marginals = marginals
        # nvar_Rn + nvar_R + nvar_P + nvar_G + pdf_R + pdf_G
        rowsize = len(marginals)*4 + 2
        # data? takový neslaný nazev...
        # data suppose to be cKDTree compatible, i.e. 
        # nsim, nvar** = data.shape
        self._data = np.empty((0, rowsize), dtype=float)
        if alpha is None:
            self.alpha = np.ones(self.nvar)
        else:
            self.set_alpha(alpha)
        
     # nemusím duplikovat a dědit 
    set_alpha = set_alpha
        
    def __repr__(self):
        return "%s(%s, %s)"%('UnCorD', repr(self.__marginals), repr(self.alpha))
        
    def __str__(f):
        return "UnCorD: " + str(f.R)
    
    def __call__(f, ns=0):  
        f_copy = copy.copy(f) # nebo deep?
        f_copy._data = np.empty((0, f_copy._data.shape[1]), dtype=float)
        
        if ns:
            sample_dict = {'P':np.random.random((ns, f_copy.nvar))}
            f._chain(sample_dict)
            pdf_G = np.prod(stats.norm.pdf(sample_dict['G']), axis=1).reshape(-1, 1)
            pdfs_R = [f.marginals[i].pdf(sample_dict['R'][:, i]) for i in range(f.nvar)]
            # je tu fakt axis=0. Dochazí totíž v iterátoru k převracení
            pdf_R = np.prod(pdfs_R, axis=0).reshape(-1, 1)
            # nvar_Rn + nvar_R + nvar_P + nvar_G + pdf_R + pdf_G
            f_copy._data = np.hstack((sample_dict['Rn'], sample_dict['R'], sample_dict['P'], sample_dict['G'], pdf_R, pdf_G))
        return f_copy
    
    def __len__(f):
        return len(f._data)
    
    def __getitem__(f, slice):
        f_copy = copy.copy(f) # nebo deep?
        f_copy._data = np.atleast_2d(f._data[slice,:]) #.copy()
        return f_copy
        
        # dúfám, že tyhle slajsy sa vyplatí
    def __getattr__(f, attr):
        if attr == 'f_model':
            return f
        elif attr == 'Rn':
            return f.__frame(0)
        elif attr == 'R':
            return f.__frame(1)
        elif attr in ('P', 'U'):
            return f.__frame(2)
        elif attr in ('GK', 'G'):
            return f.__frame(3)
            
        elif attr == 'nvar':
            return len(f.__marginals)
        elif attr == 'nsim':
            return len(f._data)
        elif attr == 'marginals':
            return f.__marginals
        elif attr == 'cor':
            return np.diag([1 for __ in range(f.nvar)])
            
        elif attr in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return getattr(f, attr[1:]) * f.alpha
            
        # hustoty
        # I'm considering to deprecate attribute access
        elif attr[:4] == 'pdf_':
            return f.pdf(attr[4:])
            
        raise AttributeError(attr)
        
    # pro určitou konzistenci. Ne že bych chtěl zamykat dveře v Pythonu
    def __setattr__(f, attr, value):
        if attr in ('_UnCorD__marginals','_data', 'alpha'):
            f.__dict__[attr] = value
        else:
            #raise AttributeError('Аслыкъёсы воштыны уг луи!')
            raise AttributeError('Atribute %s of %s object is not writable' % (attr, f.__class__.__name__))
        
    def __frame(f, i):
        nvar = f.nvar
        sl = slice(i*nvar, (i+1)*nvar)
        return f._data[:,sl]
        
        
    def add_sample(f, sample, space='R'):
        """
        Adds coordinates from input sample
        """
        newdata = f._parse_sample(sample, space)
        f._data = np.vstack((f._data, newdata))
            
        
    #č tohle už není "drobná pomucka"
    #č to je jedná z funkcí, která běží 30% času
    def new_sample(f, sample=None, space='R', extend=False):
        """
        Returns new f_model object with the same distribution and with coordinates from 'sample' taken
        """
        f_copy = copy.copy(f)
        if sample is None:
            f_copy._data = np.empty((0, f_copy._data.shape[1]), dtype=float)
        else:
            f_copy._data = np.atleast_2d(f._parse_sample(sample, space, extend))
        return f_copy
    
    
    def _parse_sample(f, input_sample, space='R', extend=False):
        # isinstance, ne?
        if f.__class__.__name__ == input_sample.__class__.__name__:
            if f.marginals == input_sample.marginals:
                if (space in ('R', 'Rn', 'P', 'GK', 'G', 'U')) or np.all(f.alpha == input_sample.alpha):
                    return input_sample._data
        
        
        # does sample is another f_model object? 
        try:
            sample_ = getattr(input_sample, space)
        except AttributeError:
            # no
            sample_ = input_sample
            
        sample_ = np.atleast_2d(sample_)
        isim, ivar = np.shape(sample_) # input sim, var
        if extend:
            to_extend = sample_
            sample_ = np.zeros((isim, f.nvar))
            sample_[:,:ivar] = to_extend
        elif ivar != f.nvar:
            raise ValueError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
            
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample_ = sample_ / f.alpha
            space = space[1:]        
                
        elif space not in ('R', 'Rn', 'P', 'GK', 'G', 'U'):
            # co jako, mám gettext sem tahnout?!
            raise ValueError('SNorm: zadaný prostor %s mi není znám' % space)
            raise ValueError('SNorm: unknown space %s' % space)
            
            
        if space=='GK':
            space='G'
        elif space=='U':
            space='P'
            
        sample_dict = {space:np.array(sample_, dtype=float).reshape(-1, f.nvar)}
        f._chain(sample_dict)
        pdf_G = np.prod(stats.norm.pdf(sample_dict['G']), axis=1).reshape(-1, 1)
        pdfs_R = [f.marginals[i].pdf(sample_dict['R'][:, i]) for i in range(f.nvar)]
        # je tu fakt axis=0. Dochazí totíž v iterátoru k převracení
        pdf_R = np.prod(pdfs_R, axis=0).reshape(-1, 1)
        # nvar_Rn + nvar_R + nvar_P + nvar_G + pdf_R + pdf_G
        newdata = np.hstack((sample_dict['Rn'], sample_dict['R'], sample_dict['P'], sample_dict['G'], pdf_R, pdf_G))
                    
        return newdata
    
    
     
    def _chain(f, sample_dict):
        # chain tam
        # чаль татысь
        if 'R' not in sample_dict and 'Rn' in sample_dict:
            sample_dict['R'] = np.empty_like(sample_dict['Rn'])
            for i in range(f.nvar):
                sample_dict['R'][:, i] = sample_dict['Rn'][:, i]*f.marginals[i].std() + f.marginals[i].mean()
                
        if 'P' not in sample_dict and 'R' in sample_dict:
            sample_dict['P'] = np.empty_like(sample_dict['R'])
            for i in range(f.nvar):
                sample_dict['P'][:, i] = f.marginals[i].cdf(sample_dict['R'][:, i])
                
        if 'G' not in sample_dict and 'P' in sample_dict:
            sample_dict['G'] = stats.norm.ppf(sample_dict['P'])
        
        # chain sem
        # чаль татчи
        elif 'P' not in sample_dict and 'G' in sample_dict:
            sample_dict['P'] = stats.norm.cdf(sample_dict['G'])
            
        if 'R' not in sample_dict and 'P' in sample_dict:
            sample_dict['R'] = np.empty_like(sample_dict['P'])
            for i in range(f.nvar):
                sample_dict['R'][:, i] = f.marginals[i].ppf(sample_dict['P'][:, i])
                
        if 'Rn' not in sample_dict and 'R' in sample_dict:
            sample_dict['Rn'] = np.empty_like(sample_dict['R'])
            for i in range(f.nvar):
                sample_dict['Rn'][:, i] = (sample_dict['R'][:, i] - f.marginals[i].mean())/f.marginals[i].std()
     
     
    def pdf(f, space='R'):
        """
        Returns own probability densities in the given space
        """
        if space == 'R':
            return f._data[:,-2]
        elif space == 'Rn':
            return f._data[:,-2] * np.prod(list(f.marginals[i].std() for i in range(f.nvar)))
        elif space in ('GK', 'G'):
            return f._data[:,-1]
        elif space in ('P', 'U'):
            return f.sample_pdf(f, space)
            
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.pdf(space[1:]) / np.prod(f.alpha)
        
        else:   
            raise ValueError('Unknown space %s' % space)
            
        
        
    def sample_pdf(f, input_sample, space='R'):
        """
        Calculates probability density for the given external sample in the given space.
        Function intended for the case no transformation needed. 
        Otherwise new f_sample should be instanciated.
        """
        
        # does sample is another f_model object? 
        # кинлы со zase кулэ?
        try:
            sample = getattr(input_sample, space)
        except AttributeError:
            # no
            sample = np.atleast_2d(input_sample)
            
        
        if space == 'R':
            pdfs_R = [f.marginals[i].pdf(sample[:, i]) for i in range(f.nvar)]
            # je tu fakt axis=0. Dochazí totíž v iterátoru k převracení
            return np.prod(pdfs_R, axis=0)
                
        elif space == 'Rn':
            sample_Rn = sample
            sample_R = np.empty_like(sample_Rn)
            for i in range(f.nvar):
                sample_R[:, i] = sample_Rn[:, i]*f.marginals[i].std() + f.marginals[i].mean()
            pdf_R = f.sample_pdf(sample_R, space='R')
            return pdf_R * np.prod(list(f.marginals[i].std() for i in range(f.nvar)))
        
        elif space in ('GK', 'G'):
            pdfs_R = stats.norm.pdf(sample)
            return np.prod(pdfs_R, axis=pdfs_R.ndim-1)
            
        elif space in ('P', 'U'):
            return np.where(np.all((sample >= 0) & (sample <= 1), axis=1), 1, 0)
            
        # new piece of code "alpha"-related
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.sample_pdf(sample / f.alpha, space=space[1:]) / np.prod(f.alpha)
            
        else:   
            raise ValueError('Unknown space %s' % space)
            





#č je чёртова прорва těch různejch pythonovejch balíků,
#č co ten Natafův model implementujou
#č takže skoro není důvod znovu to psat
#č ale udělám takovou rychlovku
#č kdyby někdo nechtěl tahnout externí závislosti
#č taky zde mám větší kontrolu nad kodem 

class Nataf: # quick'n dirty
    """
    Takes tuple of scipy stats distribution objects
    Note, there is correlation distortion in Nataf model, 
    correlation matrix meant to be in Gaussian space (just easier to implement) 
    """
    def __init__(self, marginals, gauss_cor=None, alpha=None):
        #č zatím nechcu nic kontrolovat. Python to je nebo co?
        self._marginals = marginals
        if gauss_cor is None:
            self._gauss_cor = np.diag([1 for __ in range(self.nvar)])
        else:
            self._gauss_cor = gauss_cor
        # data suppose to be cKDTree compatible, i.e. 
        # nsim, nvar = data.shape
        self.G = np.empty((0, len(marginals)), dtype=float)
        if alpha is None:
            self.alpha = np.ones(self.nvar)
        else:
            self.set_alpha(alpha)
        
     # nemusím duplikovat a dědit 
    set_alpha = set_alpha
    #č jsou ještě alfy, ty mě ale nezajimají
    #č G-čko taky ne
    _pdf_attrs = ('pdf_Rn', 'pdf_R', 'pdf_P', 'pdf_GK', 'pdf_G', 'pdf_U')
    _length_attrs = ('Rn', 'R', 'P', 'GK', 'U',\
             'pdf_Rn', 'pdf_R', 'pdf_P', 'pdf_GK', 'pdf_G', 'pdf_U')
             
    _attrs = ('_pdf_left_part', '_inv_cor', '_L', '_inv_L', '_pdf_Rn_coef')
        
    def __repr__(self):
        return "%s(%s, %s, %s)"%('Nataf', self._marginals, self._gauss_cor, self.alpha)
        
    def __str__(f):
        return "Nataf: " + str(f.R)
    
    def __call__(self, ns=0):  
        f_copy = self.__class__(self._marginals, self._gauss_cor, self.alpha)
        
        if ns:
            f_copy.G = np.random.randn(ns, len(self._marginals))
        return f_copy
    
    def __len__(f):
        return len(f.G)
    
    def __getitem__(f, slice):
        f_copy = f.__class__(f._marginals, f._gauss_cor, f.alpha)
        f_copy.G = np.atleast_2d(f.G[slice])
        for _attr in ('_Rn', '_R', '_P', '_GK', '_U'):
            try:
                setattr(f_copy, _attr, np.atleast_2d(getattr(f, _attr)[slice]))
            except:
                pass
        for attr in f._pdf_attrs:
            _attr = '_' + attr
            try:
                setattr(f_copy, _attr, np.atleast_1d(getattr(f, _attr)[slice]))
            except:
                pass
        return f_copy
        
        # dúfám, že tyhle slajsy sa vyplatí
    def __getattr__(f, attr):
        if attr == 'f_model':
            return f
        elif attr in f._length_attrs:
            return f._get_lengthy_attr('_' + attr)
        elif attr in f._attrs:
            return f._get_attr(attr)
            
        elif attr == 'nvar':
            return len(f._marginals)
        elif attr == 'nsim':
            return len(f.G)
        elif attr == 'marginals':
            return f._marginals
        elif attr == 'cor':
            return self._gauss_cor
            
        elif attr in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return getattr(f, attr[1:]) * f.alpha
            
        #č hustoty
        #č alfy. podle mě nemusí být částé
        elif attr[:4] == 'pdf_':
            return f.pdf(attr[4:])
            
        raise AttributeError(attr)
        
            
        
    #č tohle už není "drobná pomucka"
    #č to je jedná z funkcí, která běží 30% času
    def new_sample(self, sample=None, space='R', extend=False):
        """
        Returns new f_model object with the same distribution 
        and coordinates taken from the 'sample' 
        """
        f_copy = self.__class__(self._marginals, self._gauss_cor, self.alpha)
        if sample is not None:
            f_copy.add_sample(sample, space, extend)
        return f_copy
    
    
    def add_sample(f, input_sample, space='R', extend=False):
        "Adds coordinates from the input sample"
        
        # does sample is another f_model object? 
        try:
            sample_ = getattr(input_sample, space)
        except AttributeError:
            # no
            sample_ = input_sample
            
        sample_ = np.atleast_2d(sample_)
        isim, ivar = np.shape(sample_) # input sim, var
        if extend:
            to_extend = sample_
            sample_ = np.zeros((isim, f.nvar))
            sample_[:,:ivar] = to_extend
        elif ivar != f.nvar:
            raise ValueError('%sD data expected, but %sD sample given'% (f.nvar, ivar))
            
        # new piece of code "alpha"-related
        if space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            sample_ = sample_ / f.alpha
            space = space[1:]        
                
                
            
        if space=='Rn':
            f._put_attr(space, sample_)
            sample_ = f.Rn_to_R(sample_)
            space = 'R'
            
        if space=='R':
            f._put_attr(space, sample_)
            sample_ = f.R_to_P(sample_)
            space = 'P'
        
        if space=='P':
            f._put_attr(space, sample_)
            sample_ = f.P_to_GK(sample_)
            space = 'GK'
        
        if space=='GK':
            f._put_attr(space, sample_)
            sample_ = f.GK_to_G(sample_)
            space = 'G'
        
        
        if space=='U':
            f._put_attr(space, sample_)
            sample_ = f.U_to_G(sample_)
            space = 'G'
            
        
        if space=='G':
            f.G = np.vstack((f.G, sample_))
        else:
            # fail. Nothing happend...  
            #č co jako, mám gettext sem tahnout?!
            raise ValueError('Nataf: zadaný prostor %s mi není znám' % space)
            raise ValueError('Nataf: unknown space %s' % space)
    
    
    
    
    #č doufám, že se nic tu nepokazí
    def _put_attr(f, attr, value):
        "attr supposed to Rn, R, P, GK or U"
        if len(f.G):
            current = getattr(f, attr)
            setattr(f, '_' + attr, np.vstack((current, value)))
        else:
            setattr(f, '_' + attr, value)
    
    def Rn_to_R(f, Rn):
        R = np.empty_like(Rn)
        for i in range(f.nvar):
            R[:, i] = Rn[:, i]*f.marginals[i].std() + f.marginals[i].mean()
        return R
    
    def R_to_P(f, R):
        P = np.empty_like(R)
        for i in range(f.nvar):
            P[:, i] = f.marginals[i].cdf(R[:, i])
        return P
        
    def P_to_GK(f, P):
        return stats.norm.ppf(P)
        
    def GK_to_G(f, GK):
        return (f._inv_L @ GK.T).T
        
    def U_to_G(f, U):
        return stats.norm.ppf(U)
    
    #č náhrada čejnu
    def _get_lengthy_attr(f, _attr):
        """č zatím přdpokládám, že jako vstup bere něco jako _R, _Rn, _GK atd."""
        chain_function = getattr(f, '_get' + _attr)
        if _attr in f.__dict__:
            current = getattr(f, _attr)
            length = len(current)
            if length < len(f.G):
                append = chain_function(slice(length,len(f.G)))
                if length:
                    if _attr[:5] == '_pdf_':
                        current = np.hstack((current, append))
                    else:
                        current = np.vstack((current, append))
                else:
                    current = append
                setattr(f, _attr, current)
                return current
            else: #č nechce se mi dělat-robiť kontroly
                #č když je currebt z jakéhokoliv důvodu větší jak f.G
                #č tak vratím jej uživateli, ať s tím drbá won! 
                return current
        
        #č nejsem jist, jestli ta matika (tady, v čejnu) se nezlobí
        #č když ji strkám na výpočet prazdné matice
        elif len(f.G)==0:
            if _attr[:5] == '_pdf_':
                return np.empty(0) 
            else:
                return np.empty_like(f.G)
        else:
            current = chain_function(slice(0,len(f.G)))
            setattr(f, _attr, current)
            return current
    
    def _get_attr(f, current_attr):
        """č zatím přdpokládám, že jako vstup bere něco jako _L, _inv_L atd."""
        chain_function = getattr(f, '_get' + current_attr)
        #č předpokladám, že tyhle blbosti stačí spočítat jenom jednou
        current = chain_function()
        setattr(f, current_attr, current)
        return current
    
    def _get_pdf_left_part(f):
        denominator = np.power(2*np.pi, f.nvar/2) * np.sqrt(np.linalg.det(f._gauss_cor))
        return 1/denominator
            
    def _get_inv_cor(self):
        return np.linalg.inv(self._gauss_cor)
        
    def _get_L(self):
        return np.linalg.cholesky(self._gauss_cor)
        
    def _get_inv_L(self):
        return np.linalg.inv(self._L)
        
    def _get_pdf_Rn_coef(f):
        return np.prod(list(f.marginals[i].std() for i in range(f.nvar)))
            
    
    #č np.empty_like není úplně bezpěčný
    #č ať se ti nestane to, že se objeví nějaká celočiselná matice!
    #č Zatim všecho řetezí od f.G, který je explicitně s plovoucí tečkou
    #č takže zatim asi v cajku
    
    #č vidím, že numpy žere aj None jako slice. Tím lepe!
    #č žere, ale přidává dimenzi
    
#    def _get_sliced(f, attr, slice):
#       if slice is None:
#           return getattr(f, attr)
#       else:
#           return getattr(f, attr)[slice]
    
    def _get_Rn(f, slice=None):
        R = f.R[slice]
        Rn = np.empty_like(R)
        for i in range(f.nvar):
            Rn[:, i] = (R[:, i] - f.marginals[i].mean())/f.marginals[i].std()
        return Rn
            
    def _get_R(f, slice=None):
        P = f.P[slice]
        R = np.empty_like(P)
        for i in range(f.nvar):
            R[:, i] = f.marginals[i].ppf(P[:, i])
        return R
            
    def _get_P(f, slice=None):
        GK = f.GK[slice]
        P = stats.norm.cdf(GK)
        return P
    
    def _get_GK(f, slice=None):
        G = f.G[slice]
        GK = (f._L @ G.T).T
        return GK
        
    def _get_U(f, slice=None):
        G = f.G[slice]
        U = stats.norm.cdf(G)
        return U
            
            
    def _get_pdf_U(f, slice=None):
        U = f.U[slice]
        pdf_U = np.where(np.all((U >= 0) & (U <= 1), axis=1), 1, 0)
        return np.nan_to_num(pdf_U, copy=False)
        
    
    def _get_pdf_G(f, slice=None):
        G = f.G[slice]
        pdf_G = np.prod(stats.norm.pdf(G), axis=1)
        return np.nan_to_num(pdf_G, copy=False)   
        
    def _get_pdf_GK(f, slice=None):
        GK = f.GK[slice]
        right_part = np.exp(-0.5 * np.sum((GK @ f._inv_cor)*GK, axis=1))
        pdf_GK = f._pdf_left_part * right_part
        return np.nan_to_num(pdf_GK, copy=False)  
        
    def _get_pdf_P(f, slice=None):
        P = f.P[slice]
        GK = f.GK[slice]
        pdf_GK = f.pdf_GK[slice]
        pdf_GK_indep = np.prod(stats.norm.pdf(GK), axis=1)
        pdf_P = pdf_GK/pdf_GK_indep
        pdf_P = np.where(np.all((P >= 0) & (P <= 1), axis=1), pdf_P, 0)
        return np.nan_to_num(pdf_P, copy=False)  
        
    def _get_pdf_R(f, slice=None):
        pdf_P = f.pdf_P[slice]
        R = f.R[slice]
        pdfs_R = [f.marginals[i].pdf(R[:, i]) for i in range(f.nvar)]
        # je tu fakt axis=0. Dochazí totíž v iterátoru k převracení
        pdf_R = np.prod(pdfs_R, axis=0) * pdf_P
        return np.nan_to_num(pdf_R, copy=False)
        
    def _get_pdf_Rn(f, slice=None):
        pdf_R = f.pdf_R[slice]
        pdf_Rn = pdf_R * f._pdf_Rn_coef
        return np.nan_to_num(pdf_Rn, copy=False)
        
    
    def pdf(f, space='R'):
        """
        Returns own probability densities in the given space
        """
        if space in ('Rn', 'R', 'P', 'GK', 'G', 'U'):
            return f._get_lengthy_attr('_pdf_' + space)
            
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.pdf(space[1:]) / np.prod(f.alpha)
        
        else:   
            raise ValueError('Unknown space %s' % space)
            
        
        
    def sample_pdf(f, input_sample, space='R'):
        """
        Calculates probability density for the given external sample in the given space.
        Function intended for the case no transformation needed. 
        Otherwise new f_sample should be instanciated.
        """
        
        # does sample is another f_model object? 
        # кинлы со zase кулэ?
        try:
            sample = getattr(input_sample, space)
        except AttributeError:
            # no
            sample = np.atleast_2d(input_sample)
            
        
        if space == 'G':
            pdfs_R = stats.norm.pdf(sample)
            return np.prod(pdfs_R, axis=pdfs_R.ndim-1)
                
        elif space == 'U':
            return np.where(np.all((sample >= 0) & (sample <= 1), axis=1), 1, 0)
        
        elif space in ('Rn', 'R', 'P', 'GK'):
            #č tohle už není až tak triviální
            _f = f()
            _f.add_sample(sample, space=space)
            return _f.pdf(space)
            
        # new piece of code "alpha"-related
        elif space in ('aR', 'aRn', 'aGK', 'aG', 'aP', 'aU'):
            return f.sample_pdf(sample / f.alpha, space=space[1:]) / np.prod(f.alpha)
            
        else:   
            raise ValueError('Unknown space %s' % space)
            
